# High Accuracy Measurement
# ver:0.0.4
#modify:
# print("Descriptor Q Format:\t%d "%(self.ds.descriptor_q))
# modify to fit Jetson nano
#

import serial
import time
import struct
#import numpy as np

class header:
	version = ""
	totalPackLen =0
	tvlHeaderLen = 8
	platform = ""
	frameNumber = 0
	timeCpuCycles = 0
	numDetectedObj = 0
	numTLVs = 0

# Detected Objects
class dos:
	structureTag = 0
	lengthOfStruct  = 0 
	descriptor_val = 0
	descriptor_q = 0
	rangeEst_low = 0
	padding0_low = 0
	padding1_low = 0
	rangeEst_high = 0
	padding0_high = 0
	padding1_high = 0
	rangeValue = 0.0 #for range value
	
#Range Profile Infomation
class rangeProfileInfo:
	structureTag = 0
	lengthOfStruct = 0

#States Info
class stats:
	stt = 0
	los = 0
	ifpt = 0.0
	tot = 0.0
	ifpm = 0.0
	icpm = 0.0
	afcl = 0.0
	icl = .0

class HighAccuracy:
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	port = ""
	#header
	hdr = header
	#detected objects
	ds = dos()
	#range Profile
	rp = rangeProfileInfo
	#stats information
	st = stats

	#--add
	mmWaveType1 = 0 
	length1 = 0
	#descriptor_val = 0
	#descriptor_q = 0
	length2 =0
	mmWaveType6 = 0
	type6Cnt = 0
	
 
	def __init__(self,port):
		self.port = port
		print("***High Accuracy Measurement init***")
		
	
	def detectedObjectRange(self, buf):
		try:
			(self.ds.structureTag,
			self.ds.lengthOfStruct,
			self.ds.descriptor_val,
			self.ds.descriptor_q,
			self.ds.rangeEst_low,
			self.ds.padding0_low,
			self.ds.padding1_low,
			self.ds.rangeEst_high,
			self.ds.padding0_high,
			self.ds.padding1_high) = struct.unpack('2I8H', buf)
			
			self.ds.rangeValue = 0.0
		except:
			print("range error found:")
			return (False,self.ds)
		
		try:
			self.ds.rangeValue = float(int(self.ds.rangeEst_low) + int(self.ds.rangeEst_high) * 2**16) / float(2**self.ds.descriptor_q)
		except:
			print("---divide Error----")
			return (False, self.ds)
			
		#print("val={:.4f}".format(ds.rangeValue))
		return (True,self.ds)


	def getHeader(self):
		return self.hdr
		
	def getDetectedObject(self):
		return self.ds
		
	def getStatsInfo(self):
		return self.st
		
	def getRangeProfileInfo(self):
		return self.rp

	def tlvRead(self, disp):
		#print("---tlvRead---")
		 
		idx = 0
		lstate = 'idle'
		sbuf = b""
		xd = b""
		#*******Message TLV Header 1 *******
		#mmWaveType1 = 0
		#length1 = 0
		#********Message TLV Header 2******
		mmWaveType2 = 0
		length2 = 0
		length3 = 0
		mmWaveType6 = 0
		rangeProfile = b""
	
		while True:
			ch = self.port.read()
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[idx]:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					#GPIO.output(21, True)	
					idx += 1
					if idx == 8:			
						idx = 0
						lstate = 'header'
						rangeProfile = b""
						sbuf = b""		
						#print("--vital header")
				else:
					idx = 0
					return (False,self.ds, xd)
		
			elif lstate == 'header':
				sbuf += ch
				idx += 1
				if idx == 28:	
					#print("------header-----")
					#print(":".join("{:02x}".format(c) for c in sbuf))
					#print("------header end -----")  
					# [header - Magicword]
					try:   		  
						(self.hdr.version , self.hdr.totalPackLen, 
						self.hdr.platform , self.hdr.frameNumber, 
						self.hdr.timeCpuCycles, self.hdr.numDetectedObj , 
						self.hdr.numTLVs) = struct.unpack('7I', sbuf)
						
					except:
						print("Improper TLV structure found: ")
						rangeProfile = b""
						return (False,self.ds, list(xd))
						
					if disp == True:  
						print("***header***********") 
						print("PID(frame#): \t%d "%(self.hdr.frameNumber))
						print("Version:     \t%x "%(self.hdr.version))
						print("TLV:         \t%d "%(self.hdr.numTLVs))
						print("Detect Obj:  \t%d "%(self.hdr.numDetectedObj))
						print("Platform:    \t%X "%(self.hdr.platform))
						print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
						
					sbuf = b""
					idx = 0
					lstate = 'dos'
					#if length1 > 288: 
					length1 = 128 
					#print("header=>data1=>length:"%(length1))
					  
				elif idx > 28:
					idx = 0
					lstate = 'idle'
					return (False, self.ds ,list(xd))
				
			elif lstate == 'dos':
				sbuf += ch
				idx += 1
				if idx == 24:
					(tf, self.ds) = self.detectedObjectRange(sbuf)
					if disp == True:
						print("-----[Detected Objects(dos)] ----")
						print("mmWave Type 1:      \t%d "%(self.ds.structureTag))
						print("Data 1 Len:         \t%d "%(self.ds.lengthOfStruct,)) 
						print("Descriptor Value:   \t%d "%(self.ds.descriptor_val))
						print("Descriptor Q Format:\t%d "%(self.ds.descriptor_q)) 
						print("range hi:{:x}".format(self.ds.rangeEst_high))
						print("range lo:{:x}".format(self.ds.rangeEst_low))
						print("range:{:.4f}".format(self.ds.rangeValue))
						
					idx = 0
					sbuf = b""
					
					lstate = 'rpInfo'
					if not tf:
						lstate = 'idle'
						self.port.flushInput()
						return (False,self.ds, list(xd))
						 
			elif lstate == 'rpInfo':
				sbuf += ch
				idx += 1
				if idx == 8:
					#print(":".join("{:02x}".format(c) for c in sbuf))
					idx = 0 
					length2 = 0
					try:   		  
						mmWaveType2 , length2 = struct.unpack('2I',sbuf) 
						self.rp.structureTag = mmWaveType2
						self.rp.lengthOfStruct = length2
						
					except:
						print("Improper Type2, length structure found: ")
						return (False,self.ds, list(xd))
					
					lstate = 'rangeProfile'
					rangeProfile = b""
					if disp == True:
						print("[Range Profile Header] = 8 bytes")
						print("mmWave Type 2:\t%d "%(mmWaveType2))
						print("Data 2 Len:\t%d "%(length2))
						
					sbuf = b""
				  
				elif idx > 8:
					idx = 0
					lstste = 'idle'
				  
			elif lstate == 'rangeProfile':
				
				rangeProfile += ch
				idx += 1
				if idx == length2: #length2 * 8:
					#print(":".join("{:02x}".format(c) for c in rangeProfile))
					if disp == True:
						print("---------rangeProfile:" + str(len(rangeProfile)))
					try:
						#xd = struct.unpack('1024f',rangeProfile)
						xd = struct.unpack("{:d}f".format(int(length2/4)),rangeProfile)
					except:
						print("Improper rangeProfile data found: ")
						return (False,self.ds, list(xd))
        
					#GPIO.output(21, False)
					#print(list(xd))
					idx = 0 
					sbuf = b""
					lstate = "statsInfo"
					if self.type6Cnt > 2:
						lstate = 'idle'
						self.type6Cnt = 0
						if disp == True:
							print("----------re start ----------")
						self.port.flushInput()
						return (False,self.ds, list(xd))
						
				elif length2 > 4096:
					idx = 0
					sbuf = b""
					lstate = 'idle'
					print("Improper range length2 data found: ")
					return (False,self.ds, list(xd))
					
				elif idx > length2:
					idx = 0
					lstate = 'idle'
					sbuf = b""
					return (False,self.ds, list(xd))
					
			elif lstate == 'statsInfo':
				idx += 1
				sbuf += ch
				#print("statsInfo:{:d}".format(idx))
				if idx == 32:
					try:
						
						#mmWaveType6 , length3 , ifpt,tot,ifpm,icpm,afcl,icl = struct.unpack('2L6L',sbuf) 
						# For Jetson Nano => is unsigned int(I)
						mmWaveType6 , length3 , self.st.ifpt,self.st.tot,self.st.ifpm,self.st.icpm,self.st.afcl,self.st.icl = struct.unpack('2I6I',sbuf)
						self.st.stt = mmWaveType6
						self.st.los = length3
					except:
						print("Improper stats data found: ")
						return (False,self.ds, list(xd))
					
					if mmWaveType6 != 6:
						self.type6Cnt += 1
						return (False,self.ds, list(xd))
						
					if disp == True:
						print("[stats Tag/Length]")
						#print(":".join("{:02x}".format(c) for c in sbuf))
						print("mmWave Type 6:\t{:d}".format(mmWaveType6))
						print("Data 3 Len:\t{:d}".format(length3))
						print("Inter-frame Processing Time:\t{:f}".format(self.st.ifpt))
						print("Transmit Output Time:\t{:f}".format(self.st.tot))
						print("Inter-frame Processing Margin:\t{:f}".format(self.st.ifpm))
						print("Inter-chirp Processing Margin:\t{:f}".format(self.st.icpm))
						print("Active frame CPU Load:\t{:f}".format(self.st.icpm))
							
					#print(xd)
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return (True, self.ds, list(xd))
					
			elif idx > 32:
					idx = 0
					lstate = 'idle'
					sbuf = b""
					return (False,self.ds, list(xd))





# Vehicle Occupant Detection
# for only
# ver:0.0.1
# 2021/01/7
# parsing Vehicle Occupant Detection 3D data
# hardware:(Batman-201 ISK)VOD IWR6843
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: V8,V9,V10 Raw data
# v0.0.1 : 2021/01/07 release

import serial
import time
import struct
#import numpy as np

class header:
    version = 0
    totalPackLen = 0
    platform = 0
    frameNumber = 0
    timeCpuCycles = 0
    numDetectedObj = 0
    numTLVs = 0
    subFrameIndex = 0
    
class vitalSign:
    unwrapped = 0.0
    heart = 0.0
    breathing  = 0.0
    heart_rate = 0.0
    breathing_rate = 0.0

class VehicleODR:
    magicWord =  [b'\x01',b'\x02',b'\x03',b'\x04',b'\x05',b'\x06',b'\x07',b'\x08',b'\0x99']
    port = ""
    hdr = header
    vs  = vitalSign
    dck = 0
    # add for VOD interal use
    tlvLength = 0
    
    # for debug use
    dbg = False #Packet unpacket Check: True show message
    sm = False #Observed StateMachine: True Show message
    check = False # Observe numTLV and v length
    
    def __init__(self,port):
        self.port = port
        print("(jb)Vehicle Occupant Detection")
        print("(jb)For Hardware:IWR6843")
        print("(jb)SW:0.0.1: HW:IWR-6843")
        print("(jb)Firmware: V5205_VOD")
        print("(jb)UART Baud Rate:921600")
        print("(jb)==================================")
        print("(jb)Output: V8,V9,V10 data:(RAW)")
        print("(jb)V8 :Range Azimuth Heatmap TLV")
        print("(jb)V9 :Feature Vector TLV")
        print("(jb)V10:Decision Vector TLV")
        
        
    def checkTLV(self,ft):
        self.check = ft
        
    def useDebug(self,ft):
        self.dbg = ft
        
    def stateMachine(self,ft):
        self.sm = ft
        
    def getVitalSign(self):
        return self.vs
    
    def getHeader(self):
        return self.hdr
        
    def headerShow(self):
        print("***header***********")
        print("Version:     \t%x "%(self.hdr.version))
        print("Platform:    \t%X "%(self.hdr.platform))
        print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
        print("PID(frame#): \t%d "%(self.hdr.frameNumber))
        print("timeCpuCycles: \t%d "%(self.hdr.timeCpuCycles))
        print("numDetectedObj: \t%d "%(self.hdr.numDetectedObj))
        print("numTLVs: \t%d "%(self.hdr.numTLVs))
        print("subFrameIndex: \t%d "%(self.hdr.subFrameIndex))
        print("***End Of Header***")
        
    def tlvTypeInfo(self,dtype,count,dShow):
        
        dataByte = 0
        lenCount = count
        pString = ""
        stateString = "V8"
        if dtype == 8:
            dataByte= 2    #HeatMap Value Int
            pString = "Range-Azimuth Heat Map"
            lenCount = count
            stateString = 'V8'
        elif dtype == 9:
            lenCount = count
            dataByte = 20  #target struct 20 bytes:(avgPower1,avgPower2,powerRatio1,powerRatio2,crossCorr)
            pString = "Feature Vector TLV"
            stateString = "V9"
        elif dtype == 10:
            lenCount = count
            dataByte = 12 #zone = 1 byte
            pString = "Decision Vector TLV"
            stateString = "V10"
            
        else:
            pString = "*** Type Error ***"
            stateString = 'idle'

        if dShow == True:
            print("-----[{:}] ----".format(pString))
            print("tlv Type :  \t{:d}".format(dtype))
            print("tlv length:      \t{:d}".format(lenCount))
            #print("{:}      \t{:d}".format(nString,int(nPoint)))
            #print("value length:    \t{:d}".format(retCnt))
        
        return stateString,dataByte,lenCount,pString
        
#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#            False: hide printing message
# output:(return parameter)
# (pass_fail, v8, v9, v10)
#  pass_fail: True: Data available    False: Data not available
#
#    Output: V8,V9,V10 data:(RAW)")
#    V8 :Range Azimuth Heatmap TLV
#    V9 :Feature Vector TLV
#    V10:Decision Vector TLV
#

    def tlvRead(self,disp):
        #print("---tlvRead---")
        #ds = dos
        typeList = [8,9,10]
        idx = 0
        lstate = 'idle'
        sbuf = b""
        lenCount = 0
        unitByteCount = 0
        dataBytes = 0
        
        tlvCount = 0
        pbyte = 16
        v8 = []
        v9 = ([])
        v10 = ([])
        
        zone = 0
    
        while True:
            try:
                ch = self.port.read()
            except:
                return (False,v8,v9,v10)
            #print(str(ch))
            if lstate == 'idle':
                #print(self.magicWord)
                if ch == self.magicWord[idx]:
                    #print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
                    idx += 1
                    if idx == 8:
                        #print(self.magicWord)
                        idx = 0
                        lstate = 'header'
                        rangeProfile = b""
                        sbuf = b""
                else:
                    #print("not: magicWord state:")
                    idx = 0
                    rangeProfile = b""
                    return (False,v8,v9,v10)
        
            elif lstate == 'header':
                sbuf += ch
                idx += 1
                if idx == 24:
                    #print("------header-----")
                    #print(":".join("{:02x}".format(c) for c in sbuf))
                    #print("len:{:d}".format(len(sbuf)))
                    # [header - Magicword]
                    try:
                        #(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,
                        #self.hdr.frameNumber,self.hdr.timeCpuCycles,self.hdr.numDetectedObj,
                        #self.hdr.numTLVs,self.hdr.subFrameIndex) = struct.unpack('8I', sbuf)
                        (self.hdr.totalPackLen,self.hdr.platform,
                        self.hdr.frameNumber,self.hdr.timeCpuCycles,self.hdr.numDetectedObj,
                        self.hdr.numTLVs) = struct.unpack('6I', sbuf)
                         
                    except:
                        if self.dbg == True:
                            print("(Header)Improper TLV structure found: ")
                        return (False,v8,v9,v10)
                    
                    if disp == True:
                        self.headerShow()
                    
                    tlvCount = self.hdr.numTLVs
                    if self.hdr.numTLVs == 0:
                        return (False,v8,v9,v10)
                        
                    if self.sm == True:
                        print("(Header)")
                        
                    sbuf = b""
                    idx = 0
                    lstate = 'TL'
                    dck = 0
                      
                elif idx > 40:
                    idx = 0
                    lstate = 'idle'
                    return (False,v8,v9,v10)
                    
            elif lstate == 'TL': #TLV Header type/length
                sbuf += ch
                idx += 1
                if idx == 8:
                    #print(":".join("{:02x}".format(c) for c in sbuf))
                    try:
                        ttype,self.tlvLength = struct.unpack('2I', sbuf)
                        if self.check:
                            print("(check) numTLVs({:d}): tlvCount({:d})-------ttype:tlvLength:{:d}:{:d}".format(self.hdr.numTLVs,tlvCount,ttype,self.tlvLength))
                        if ttype not in typeList or self.tlvLength > 10000:
                            if self.dbg == True:
                                print("(TL)Improper TL Length(hex):(T){:d} (L){:x} numTLVs:{:d}".format(ttype,self.tlvLength,self.hdr.numTLVs))
                            sbuf = b""
                            idx = 0
                            lstate = 'idle'
                            self.port.flushInput()
                            return (False,v8,v9,v10)
                            
                    except:
                        if self.dbg == True:
                            print("TL unpack Improper Data Found:")
                        self.port.flushInput()
                        return (False,v8,v9,v10)
                    
                    lstate,dataBytes,lenCount,pString = self.tlvTypeInfo(ttype,self.tlvLength,disp)
                    if lstate == 'V10':
                        self.zone = self.tlvLength
                     
                    if self.sm == True:
                        print("(TL:{:d})=>({:})".format(tlvCount,lstate))
                        
                    tlvCount -= 1
                    idx = 0
                    sbuf = b""
            
                    
            elif lstate == 'V8': # count = Total Lentgh - 8
                sbuf += ch
                idx += 1
                if (idx%dataBytes == 0):
                    try:
                        #print(":".join("{:02x}".format(c) for c in sbuf))
                        v8s = struct.unpack('h', sbuf)
                        v8.append(v8s[0])
                        sbuf = b""
                    except:
                        if self.dbg == True:
                            print("(6.1)Improper Type V8 structure found: ")
                        return (False,v8,v9,v10)
                    
                if idx == lenCount:
                    if disp == True:
                        print("v8[{:d}]".format(len(v8)))
                    idx = 0
                    sbuf = b""
                    if tlvCount <= 0: # Back to idle
                        lstate = 'idle'
                        dck += 1
                        if self.sm == True:
                            print("(V8:{:d})=>(idle) :true".format(tlvCount))
                        return (True,v8,v9,v10)
                        
                    else: # Go to TL to get others type value
                        lstate = 'TL'
                        dck += 1
                        idx = 0
                        sbuf = b""
                        #print("---------v8 ok ---------{:d}".format(len(v8)))
                        if self.sm == True:
                            print("(V8:{:d})=>(TL)".format(tlvCount))
                    
                elif idx > 7000: #lenCount:
                    print("V8 data over:10")
                    idx = 0
                    sbuf = b""
                    lstate = 'idle'
                    return (False,v8,v9,v10)
                    
            elif lstate == 'V9':
                idx += 1
                sbuf += ch
                if  (idx%dataBytes == 0):
                    #print("*************v9******:{:}   tlvCount={:}".format(lenCount,tlvCount))
                    #(avgPower1,avgPower2,powerRatio1,powerRatio1,crossCorr) = struct.unpack('5f',sbuf)
                    #v9.append((avgPower1,avgPower2,powerRatio1,powerRatio1,crossCorr)) # =  struct.unpack('5f',sbuf)
                    #(avgPower1,avgPower2,powerRatio1,powerRatio1,crossCorr) = struct.unpack('5f',sbuf)
                    v9.append(struct.unpack('5f',sbuf))
                    
                    
                    sbuf = b""
                    if disp == True:
                        print("v9[{:d}]".format(len(v9)))
                            
                if idx == lenCount:
                    if disp == True:
                        print("v9[{:d}]".format(len(v9)))
                    idx = 0
                    sbuf = b""
                    if tlvCount <= 0: # Back to idle
                        lstate = 'idle'
                        if self.sm == True:
                            print("(V9:{:d})=>(idle) :true".format(tlvCount))
                        return (True,v8,v9,v10)
                        
                    else: # Go to TL to get others type value
                        lstate = 'TL'
                        if self.sm == True:
                            print("(V9:{:d})=>(TL)".format(tlvCount))
                        
                if idx > lenCount :
                    print("V9 data over:80")
                    idx = 0
                    lstate = 'idle'
                    sbuf = b""
                    if self.sm == True:
                        print("(V9)=>(idle)")
                    return (False,v8,v9,v10)
                
            elif lstate == 'V10':
                idx += 1
                sbuf += ch
                #print(idx)
                if  (idx%dataBytes == 0):
                    #print("V10:dataBytes({:d}) lenCount({:d}) index:{:d}".format(dataBytes,lenCount,idx))
                    (percent,power,rangeIdx,azimuthIdx) =  struct.unpack('2f2h',sbuf)
                    v10.append((percent,power,rangeIdx,azimuthIdx))
                    
                    if disp == True:
                        print("v10[{:d}]".format(len(v10)))
                    sbuf = b""
                            
                if idx == lenCount:
                    if disp == True:
                        print("v10[{:d}]".format(len(v10)))
                    idx = 0
                    sbuf = b""
                    if tlvCount <= 0: # Back to idle
                        lstate = 'idle'
                        if self.sm == True:
                            print("(V10:{:d})=>(idle) :true".format(tlvCount))
                        return (True,v8,v9,v10)
                        
                    else: # Go to TL to get others type value
                        lstate = 'TL'
                        if self.sm == True:
                            print("(V10:{:d})=>(TL)".format(tlvCount))
                        
                if idx > lenCount :
                    print("V10 data over:96")
                    idx = 0
                    lstate = 'idle'
                    sbuf = b""
                    if self.sm == True:
                        print("(V10)=>(idle)")
                    return (False,v8,v9,v10)





# Traffic Monitoring Detection Roadway Sensing key Value data
# ver:0.1.0
# create date: 2021/04/21
# parsing Roadway Traffic Monitoring Detect
# hardware:(BM-201 TRS): ISK IWR6843 ES2.0
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: objPoint: ['flow','fn','indexMax','index','x','y','range','doppler','area','ptsNum','cid']
# v0.1.0 : 2021/04/21 release
# v0.1.1 : 2021/05/02 release
# v0.1.2 : 2021/05/04 release (bug fix)
#	Add:	getRecordData(self,frameNum):
#			readFile(self,fileName):
#
 

import serial
import time
import struct
import pandas as pd
import numpy as np
from dataclasses import dataclass


class header:
	version = 'v0.1.1'
	frameNumber = 0


class roadwayTmdISK_kv:
	#				{        }        ,      ;
	#magicWord =  [b'\x7B',b'\x7D',b'\x3C',b'\x3E']
	magicWord =  [b'{',b'}',b'J',b'B']
	hdr = header
	frameNumber = 0
	port = ""
	v21_col_names_rt = ['flow','fn','indexMax','index','x','y','range','doppler','area','ptsNum','cid']
	v21_col_names_file = ['fn','indexMax','index','x','y','range','doppler','area','ptsNum','NotObject','MAN','MotorCycle','car','CAR']
	
	sim_startFN = 0
	sim_stopFN  = 0 
	v21simo = []
	
	def __init__(self,port):
		self.port = port
		print("(jb)Traffic Monitoring Detection Roadway Sensing lib initial")
		print("(jb)For Hardware: Batman-201(ISK)")
		print("(jb)Hardware: IWR-6843 ES2.0")
		print("(jb)Firmware: TRS_kv")
		print("(jb)UART Baud Rate:921600")

		print("(jb)Data type: kv (DataFrame)")
		print("==============Info=================")
		print("Output Data Type: DataFrame")
		print("Output: ['flow','fn','indexMax','index','x','y','range','doppler','area','ptsNum','cid']")
		print("===================================")
	
	def list2df(self,dck,l21):
		ld21 = pd.DataFrame(l21,columns=self.v21_col_names_rt)
		return (dck,ld21)
	
	def getHeader(self):
		return self.hdr
	
	def trsRead(self,disp):
		
		frameNum = 0
		idx = 0
		lstate = 'idle'
		sbuf = b""
		v21df = ([])
		while True:
			try:
				ch = self.port.read()
			except:
				print("(TRS)---port.read() Exception---")
				return self.list2df(False,v21df)
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[0]:
					#if disp:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					idx = 0
					sbuf = b""
					lstate = 'iData'
					v21df = ([])
					sbuf = ch
					#if disp:
					#	print("-----------------------")
					#	print("(jb) idle-> idata")
					
				else:
					idx = 0
					sbuf = b""
					v21df = ([])
					return self.list2df(False,v21df)
					
			elif lstate == 'iNext':
				sbuf += ch
				idx += 1
				#print(":".join("{:02x}".format(c) for c in sbuf))  
				if self.magicWord[0] == ch:
					#print(":".join("{:02x}".format(c) for c in sbuf))   
					#print("(jb)iTarget_end state:")
					lstate = 'iData'
				
				elif idx > 44:
					lstate = 'idle'
					sbuf = b""
					print("*********  data over(iNext) *********")
					print("Please close the other process in use")
		
			elif lstate == 'iData':
				sbuf += ch
				#print(":".join("{:02x}".format(c) for c in sbuf))  
				idx += 1
				if  self.magicWord[1] == ch and len(sbuf) > 44: # } 
					#print(":".join("{:02x}".format(c) for c in sbuf))
					if disp:
						print("------rx data(iData)-----")
						print(":".join("{:02x}".format(c) for c in sbuf)) 
						
					try:
						 
						(h,j,b,flow,fn,indexMax,index,x,y,ran,doppler,area,ptsNum,cid,t) = struct.unpack('3cBI9fc',sbuf[0:45])
						frameNum = fn
						v21df.append((flow,fn,indexMax,index,x,y,ran,doppler,area,ptsNum,cid))
						
						compIdx = 0 if (indexMax - 1) < 0 else indexMax - 1
						if index == compIdx:
							lstate = 'idle'
							self.hdr.frameNumber = frameNum
							#print("lib:frameNumber:{:}".format(self.hdr.frameNumber))
							v21pd = v21df
							v21df =([]) 
							return self.list2df(True,v21pd) 
						else:
							lstate = 'iNext'
							
						sbuf = b""
						idx = 0
						
						if disp:
							print("(jb)iData(1) -> idle state")
			
					except:
						lstate = 'idle'
						
						print("(jb)---iData Exception---:{:}".format())
						if disp:
							print("(jb)---iData Exception---")
					
				elif idx > 45:
					lstate = 'idle'
					sbuf = b""
					print("*********data over(iData)*********")
					if disp:
						print("data over {:d} back to idle state".format(idx))
					
					v21df = ([])
					idx = 0
					
					return self.list2df(False,v21df) 
					
	#================== For playback method ======================================
	# add in v0.1.1
	#================== output data based on frameNum ==================
	def getRecordData(self,frameNum):
		s_fn = frameNum + self.sim_startFN
		#print("====getRecordData():fn:{:}".format(s_fn))
		v21d = self.v21simo[self.v21simo['fn'] == float(s_fn)]
		chk = 0
		if v21d.count != 0:
			chk = 1
		return (chk,v21d)
	#================== read data from file for use
	def readFile(self,fileName):
		
		self.fileName = fileName 
		#df = pd.read_csv(self.fileName, names = self.v21_col_names_file, dtype={'fn': float,'indexMax': float,'index':float,'x':float,'y':float,'range':float,'doppler':float,'area':float,'ptsNum':float,'cid':float}) 
		df = pd.read_csv(self.fileName, 
			names = self.v21_col_names_file,
			skiprows = 1, #delete header
			error_bad_lines=False,
			dtype={'fn': float,'indexMax':float,'index':float,'x':float,'y':float,'range':float,'doppler':float,'area':float,'ptsNum':float,'NotObject':int,'MAN':int,'MotorCycle':int,'car':int,'CAR':int}) 
		
		self.sim_startFN = df['fn'].values[0]
		df.dropna()
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		self.v21simo = df
		return df



# Overhead People Counting 3D - ISK & AOP Raw Data (OHPC)
# ver:2.0
# 2022/04/19
# 
# parsing People Counting 3D data use AOP-ISK
# hardware:(Batman-201)ISK IWR6843 ES2.0 & BM-501 AOP
#    Fireware version: I470
# Fit : wall mount
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: V6,V7,V8 Raw data
# v0.0.1 : 2020/06/19 release
#          (1)Output list data
# v0.1.0 : 2021/04/01 
#          (1)Output DataFrame
# v0.1.1 : 2021/08/12
#           (doppler and range swap in V6 output)
# v0.1.2 : 2021/09/17
#           v7 dataframe:'tid' move to front of 'ec0'
# v2.0 : 2022/04/19
#          pc3_v2: add ec[16],g,confidenceLevel
#          this version lib same as pc3OVH 
#
# v2.0.1: 2022/06/10 v6 add sx,sy,sz
#
# v2.0.2: 2022/06/13 added azimuth offset
#                    azi_offset
#

import serial
import time
import struct
import pandas as pd
import numpy as np

class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	subframeNumber = 0
	chirpMargin = 0
	frameMargin = 0 
	trackProcessTime = 0
	uartSendTime = 0
	numTLVs = 0
	checksum = 0

class unitS:
	elevationUnit:float = 0.0
	azimuthUnit : float = 0.0
	dopplerUnit :float = 0.0
	rangeUnit :float = 0.0
	snrUnit :float = 0.0

class Pc3_v2:
	
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	port = ""
	hdr = header
	u = unitS
	frameNumber = 0
	azi_offset = 0.0
	
	# provide csv file dataframe
	# real-time 
	v6_col_names_rt = ['fN','type','elv','azimuth','range','doppler','snr','sx', 'sy', 'sz'] 
	v7_col_names_rt = ['fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi','tid']
	v8_col_names_rt = ['fN','type','targetID']
	
	# read from file for trace point clouds
	fileName = ''
	v6_col_names = ['time','fN','type','elv','azimuth','range','doppler','snr','sx', 'sy', 'sz']
	#v7_col_names = ['time','fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi','tid']
	v7_col_names = ['time','fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','tid','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi'] #v0.1.2
	v8_col_names = ['time','fN','type','targetID']
	sim_startFN = 0
	sim_stopFN  = 0 
	
	# add for interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm  = False #Observed StateMachine: True Show message
	plen = 16 
	
	def __init__(self,port, degree = None):
		self.port = port
		self.azi_offset = 0.0 if degree == None else degree * 3.14159/180.0
		
		print("(jb)People Counting 3D v2.0 (Wall Mount) initial")
		print("(jb)vsersion:v2.0")
		print("(jb)For Hardware:Batman-201(ISK) & BM-501 (AOP)")
		print("(jb)Hardware: IWR-6843 & AOP")
		print("(jb)Firmware: PC3-I470")
		print("(jb)UART Baud Rate:921600")
		print("Output: V6,V7,V8 data:(RAW)")
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	def headerShow(self):
		print("******* Header ********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("Inter-frame Processing Time:\t{:d} us".format(self.hdr.trackProcessTime))
		print("UART Send Time:\t{:d} us".format(self.hdr.uartSendTime))
		print("Inter-chirp Processing Margin:\t{:d} us".format(self.hdr.chirpMargin))
		print("Inter-frame Processing Margin:\t{:d} us".format(self.hdr.frameMargin))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("Check Sum   :\t{:x}".format(self.hdr.checksum))
		print("***End Of Header***") 
			
	#for class internal use
	def tlvTypeInfo(self,dtype,count,dShow):
		
		sbyte = 8  #tlvHeader Struct = 8 bytes
		unitByte = 20 
		dataByte = 0
		pString = ""
		nString = "numOfPoints :"
		stateString = "V6-unit"
		if dtype == 6:
			unitByte = 20  #pointUnit= 20bytes (elvUnit(4),aziUnit(4),dopplerUnit(4),rangeUnit(4),snrUnit(4))
			sbyte = 8      #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 8    #pointStruct 8bytes:(elevation(1),azimuth(1),doppler(2),range(2),snr(2))
			pString = "Point Cloud TLV and pointUnit"
			nString = ""
		elif dtype == 7:
			unitByte = 0   #pointUnit= 0bytes 
			sbyte = 8 	   #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte = 112  #target struct 112 bytes:(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi)  
			pString = "Target Object TLV"
			nString = "numOfObjects:"
			stateString = "V7"
		elif dtype == 8:
			unitByte = 0 #pointUnit= 0bytes 
			sbyte = 8    #tlvHeader Struct = 8 bytes
			dataByte = 1 #targetID = 1 byte
			pString = "Target Index TLV"
			nString = "numOfIDs"
			stateString = "V8"
		else:
			unitByte = 0
			sbyte = 1
			pString = "*** Type Error ***"
			stateString = 'idle'
		 
		retCnt = count - unitByte -sbyte
		nPoint = retCnt / sbyte
		#dShow = True
		if dShow == True:
			print("-----[{:}] ----".format(pString))
			print("tlv Type({:2d}B):  \t{:d}".format(sbyte,dtype))
			print("tlv length:      \t{:d}".format(count)) 
			print("{:}      \t{:d}".format(nString,int(nPoint)))
			print("value length:    \t{:d}".format(retCnt))  
		
		return unitByte,stateString, sbyte, dataByte,retCnt, int(nPoint)
		
	def list2df(self,dck,l6,l7,l8):
		ll6 = pd.DataFrame(l6,columns=self.v6_col_names_rt)
		ll7 = pd.DataFrame(l7,columns=self.v7_col_names_rt)
		ll8 = l8 #pd.DataFrame(l8,columns=self.v8_col_names_rt)
		return (dck,ll6,ll7,ll8)

#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
#     df: 
# output:(return parameter)
# (pass_fail, v6, v7, v8)
#  pass_fail: True: Data available    False: Data not available
#  v6: point cloud infomation
#  v7: Target Object information
#  v8: Target Index information
#  
# 
#
	def tlvRead(self,disp,df = None):
		
		#print("---tlvRead---")
		#ds = dos
		typeList = [6,7,8]
		idx = 0
		lstate = 'idle'
		sbuf = b""
		lenCount = 0
		unitByteCount = 0
		dataBytes = 0
		numOfPoints = 0
		tlvCount = 0
		pbyte = 16
		v6 = ([])
		v7 = ([])
		v8 = ([])
		v6df = ([])
		v7df = ([])
		v8df = ([])
		
		while True:
			try:
				ch = self.port.read()
			except:
				#return (False,v6,v7,v8)
				#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6,v7,v8)
				return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
				
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[idx]:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					idx += 1
					if idx == 8:
						idx = 0
						lstate = 'header'
						rangeProfile = b""
						sbuf = b""
				else:
					#print("not: magicWord state:")
					idx = 0
					rangeProfile = b""
					#return (False,v6,v7,v8)
					#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
		
			elif lstate == 'header':
				sbuf += ch
				idx += 1
				if idx == 40: 
					#print("------header-----")
					#print(":".join("{:02x}".format(c) for c in sbuf)) 	 
					#print("len:{:d}".format(len(sbuf))) 
					# [header - Magicword]
					try: 
						(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,
						self.hdr.frameNumber,self.hdr.subframeNumber,
						self.hdr.chirpMargin,self.hdr.frameMargin,self.hdr.trackProcessTime,self.hdr.uartSendTime,
						self.hdr.numTLVs,self.hdr.checksum) = struct.unpack('9I2H', sbuf)
						self.frameNumber = self.hdr.frameNumber
					except:
						if self.dbg == True:
							print("(Header)Improper TLV structure found: ")
						#return (False,v6,v7,v8)
						#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
					if disp == True:  
						self.headerShow()
					
					tlvCount = self.hdr.numTLVs
					#print("tlvCount:{:}".format(tlvCount))
					if self.hdr.numTLVs == 0:
						#return (True,v6,v7,v8)
						#return (True,v6,v7,v8) if (df == None) else self.list2df(True,v6df,v7df,v8df)
						return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
						
					if self.sm == True:
						print("(Header)")
						
					sbuf = b""
					idx = 0
					lstate = 'TL'
					  
				elif idx > 44:
					idx = 0
					lstate = 'idle'
					#return (False,v6,v7,v8)
					#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
			elif lstate == 'TL': #TLV Header type/length
				sbuf += ch
				idx += 1
				if idx == 8:
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						ttype,self.tlvLength = struct.unpack('2I', sbuf)
						#print("--tlvNum:{:d}: tlvCount({:d})-------ttype:tlvLength:{:d}:{:d}".format(self.hdr.numTLVs,tlvCount,ttype,self.tlvLength))
						if ttype not in typeList or self.tlvLength > 10000:
							if self.dbg == True:
								print("(TL)Improper TL Length(hex):(T){:d} (L){:x} numTLVs:{:d}".format(ttype,self.tlvLength,self.hdr.numTLVs))
							sbuf = b""
							idx = 0
							lstate = 'idle'
							self.port.flushInput()
							#return (False,v6,v7,v8)
							#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
							return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
							
					except:
						if self.dbg == True:
							print("TL unpack Improper Data Found:")
						self.port.flushInput()
						#return (False,v6,v7,v8)
						#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
					unitByteCount,lstate ,plen ,dataBytes,lenCount, numOfPoints = self.tlvTypeInfo(ttype,self.tlvLength,disp)
					#if ttype == 6:
					#	print("--pointCloud:((tlvLength({:d})-pointUnit(20)-tlvStruct(8))/8={:d}".format(self.tlvLength,numOfPoints))
					if self.sm == True:
						print("(TL:{:d})=>({:})".format(tlvCount,lstate))
						
					tlvCount -= 1
					idx = 0  
					sbuf = b""
			
			elif lstate == 'V6-unit':
				sbuf += ch
				idx += 1
				if idx == unitByteCount :
					#print(":".join("{:02x}".format(c) for c in sbuf))
					#print("unitByte:{:d}".format(len(sbuf)))
					try:
						self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit = struct.unpack('5f', sbuf)
						#print("Unit  ==> elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit))
						sbuf = b""
						idx = 0
						lstate = 'V6'				
					except:
						if self.dbg == True:
							print("(6.0)Improper Type 6 unit Value structure found: ")
						#return (False,v6,v7,v8)
						#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
					if self.sm == True:
						print("(V6-unit:{:d})=>({:})".format(tlvCount,lstate))
						#print("(V6-unit)=>({:})".format(lstate))
					
			elif lstate == 'V6': # count = Total Lentgh - 8
				sbuf += ch
				idx += 1
				if (idx%dataBytes == 0):
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						(e,a,d,r,s) = struct.unpack('2b3h', sbuf)
						elv = e * self.u.elevationUnit
						azi = a * self.u.azimuthUnit + self.azi_offset
						dop = d * self.u.dopplerUnit
						ran = r * self.u.rangeUnit
						snr = s * self.u.snrUnit
						#print("({:2d}:{:4d})(idx:({:4d}) elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(numOfPoints,lenCount,idx,elv,azi,dop,ran,snr))
						sz  = ran * np.sin(elv)
						sx  = ran * np.cos(elv) * np.sin(azi)
						sy  = ran * np.cos(elv) * np.cos(azi)
						
						if (df == 'DataFrame'):
							#v6df.append((self.hdr.frameNumber,'v6',elv,azi,dop,ran,snr,sx,sy,sz)) #v0.1.0
							v6df.append((self.hdr.frameNumber,'v6',elv,azi,ran,dop,snr,sx,sy,sz)) #v0.1.1
							 
						else:
							#v6.append((elv,azi,dop,ran,snr)) 
							v6.append((elv,azi,dop,ran,snr,sx,sy,sz)) #v2.0.1
						
						#print("point_cloud_2d.append:[{:d}]".format(len(point_cloud_2d)))
						sbuf = b""
						
					except:
						if self.dbg == True:
							print("(6.1)Improper Type 6 Value structure found: ")
						#return (False,v6,v7,v8)
						#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
				if idx == lenCount:
					if disp == True:
						print("v6[{:d}]".format(len(v6)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V6:{:d})=>(idle) :true".format(tlvCount))
						#return (True,v6,v7,v8)
						#return (True,v6,v7,v8) if (df == None) else self.list2df(True,v6df,v7df,v8df)
						return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
						
					else: # Go to TL to get others type value
						lstate = 'TL' #'tlTL'
						if self.sm == True:
							print("(V6:{:d})=>(TL)".format(tlvCount))
					
				elif idx > lenCount:
					idx = 0
					sbuf = b""
					lstate = 'idle'
					#return (False,v6,v7,v8)
					#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6,v7,v8)
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
				
			elif lstate == 'V7':
				sbuf += ch
				idx += 1
				if (idx%dataBytes == 0):
					#print("V7:dataBytes({:d}) lenCount({:d}) index:{:d}".format(dataBytes,lenCount,idx))
					try:
						#(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ) = struct.unpack('I9f', sbuf)
						(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi) = struct.unpack('I27f', sbuf)
						if (df == 'DataFrame'): 
							v7df.append((self.hdr.frameNumber,'v7',posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi,tid))
							
						else:
							v7.append((tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi))
						
						sbuf = b""
					except:
						if self.dbg == True:
							print("(7)Improper Type 7 Value structure found: ")
						#return (False,v6,v7,v8)
						#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
						
				if idx >= lenCount:
					if disp == True:
						print("v7[{:d}]".format(len(v7)))
					 
					sbuf = b""
					if tlvCount == 0:
						lstate = 'idle'
						if self.sm == True:
							print("(V7)=>(idle) :true")
						#return (True,v6,v7,v8)
						#return (True,v6,v7,v8) if (df == None) else self.list2df(True,v6df,v7df,v8df)
						return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						idx = 0
						if self.sm == True:
							print("(V7)=>(TL)")

				if idx > lenCount:
					idx = 0 
					lstate = 'idle'
					sbuf = b""
					#return (False,v6,v7,v8)
					#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
				
			elif lstate == 'V8':
				idx += 1
				v8.append(ord(ch))
				
				if idx == lenCount:
					if disp == True:
						print("=====V8 End====")						
					sbuf = b""
					idx = 0
					lstate = 'idle'
					if self.sm == True:
						print("(V8:{:d})=>(idle)".format(tlvCount))
					
					if (df == 'DataFrame'):
						v8df = [self.hdr.frameNumber,'v8']
						v8df.extend(v8)
						
					#return (True,v6,v7,v8) if (df == None) else self.list2df(True,v6df,v7df,v8df)
					return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
				
				if idx > lenCount:
					sbuf = b""
					idx = 0
					lstate = 'idle'
					#return (False,v6,v7,v8)
					#return (False,v6,v7,v8) if (df == None) else self.list2df(False,v6df,v7df,v8df)
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
	def v67Simlog(frameNum):
		global sim_startFN,sim_stopFN
		s_fn = frameNum + sim_startFN
		#print("frame number:{:}".format(s_fn))
		v6d = v6sim[v6sim['fN'] == s_fn]
		#v6d =  v6dd[v6dd['doppler'] < 0.0]
		#print(v6d)
		v7d = v7sim[v7sim['fN'] == s_fn]
		chk = 0
		if v6d.count != 0:
			chk = 1
		return (chk,v6d,v7d)
		
	def getRecordData(self,frameNum):
		s_fn = frameNum + self.sim_startFN
		#print("frame number:{:}".format(s_fn))
		v6d = self.v6simo[self.v6simo['fN'] == s_fn]
		#v6d =  v6dd[v6dd['doppler'] < 0.0]
		#print(v6d)
		v7d = self.v7simo[self.v7simo['fN'] == s_fn]
		v8d = self.v8simo[self.v8simo['fN'] == s_fn]
		chk = 0
		if v6d.count != 0:
			chk = 1
		return (chk,v6d,v7d,v8d)
		
	def readFile(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		#          ['time','fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz']
		df = pd.read_csv(self.fileName, names = self.v6_col_names, skiprows = [0,11,12]) 
		df.dropna()
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		print(df.info())
		print(df.info(memory_usage="deep")) 
		
		v6simOri = df[(df.type == 'v6')]
		#print("-------------------v6sim------------:{:}".format(v6simOri.shape))
		#self.v6simo = v6simOri.loc[:,['fN','elv','azimuth','range','doppler','snr']]
		self.v6simo = v6simOri.loc[:,['fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz']]
		self.v6simo['elv'] = self.v6simo['elv'].astype(float, errors = 'raise') 
		
		df7 = pd.read_csv(self.fileName, names = self.v7_col_names, skiprows = [0])  
		
		v7simc = df7[df7['type'] == 'v7']
		self.v7simo  = v7simc.loc[:,['fN','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi','tid']]
		self.sim_startFN = df['fN'].values[0]
		self.sim_stopFN  = df['fN'].values[-1]
		'''
		print("---------start frame number---------:{:}".format(self.sim_startFN))
		print("---------stop  frame number---------:{:}".format(self.sim_stopFN))
		print(self.v7simo)
		print("-----v8sim data lib-----")
		'''
		v8simc = df[df['type'] == 'v8']
		self.v8simo  = v8simc.loc[:,['fN','type','elv']]
		self.v8simo.columns = ['fN','type','targetID']
		#print(self.v8simo)
		return (self.v6simo,self.v7simo,self.v8simo)





# vitalsign
# ver:0.0.10
# 2020/02/04
#
#
import serial
#import time
import struct
#import numpy as np

class vsos:
	rangeBinIndexMax  = 0 
	rangeBinIndexPhase = 0 
	maxVal  = float(0.0)
	processingCyclesOut =  0 
	processingCyclesOut1 =  0 
	rangeBinStartIndex  =  0 
	rangeBinEndIndex    =  0  
	unwrapPhasePeak_mm  = float(0.0)
	outputFilterBreathOut = float(0.0)
	outputFilterHeartOut = float(0.0)
	heartRateEst_FFT     = float(0.0)
	heartRateEst_FFT_4Hz  = float(0.0)
	heartRateEst_xCorr   = float(0.0)
	heartRateEst_peakCount  = float(0.0)
	breathingRateEst_FFT   = float(0.0)
	breathingEst_xCorr     = float(0.0)
	breathingEst_peakCount  = float(0.0)
	confidenceMetricBreathOut  = float(0.0)
	confidenceMetricBreathOut_xCorr  = float(0.0)
	confidenceMetricHeartOut   = float(0.0)
	confidenceMetricHeartOut_4Hz  = float(0.0)
	confidenceMetricHeartOut_xCorr  = float(0.0)
	sumEnergyBreathWfm = float(0.0)
	sumEnergyHeartWfm  = float(0.0)
	motionDetectedFlag = float(0.0)
	rsv  = [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
	
class header:
	version = ""
	totalPackLen =0
	tvlHeaderLen = 8
	platform = ""
	frameNumber = 0
	timeCpuCycles = 0
	numDetectedObj = 0
	numTLVs = 0
	rsv = 0
	
	
class VitalSign:
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	port = ""
	
	hdr = header
	vs = vsos()
	'''
	version = ""
	totalPackLen =0
	tvlHeaderLen = 8
	platform = ""
	frameNumber = 0
	timeCpuCycles = 0
	numDetectedObj = 0
	numTLVs = 0
	'''
		
	def __init__(self,port):
		self.port = port
		print("***vital sign init***")
		
	def vital_port(self):
		print("------vital Sign ---- ok:" + self.port)
		
		
	def showHeader(self):
		print("**********************") 
		print("PID:\t%d "%(self.hdr.frameNumber))
		print("Version:\t%x "%(self.hdr.version))
		print("TLV:\t\t%d "%(self.hdr.numTLVs))
		print("Detect Obj:\t%d "%(self.hdr.numDetectedObj))
		print("Platform:\t%X "%(self.hdr.platform))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
	
	def getvitalSignsOutputStats(self):
		return self.vs
		
	def getHeader(self):
		return self.hdr

	def vitalSignsOutputStats(self,buf):
		
		try:
			(self.vs.rangeBinIndexMax,
			self.vs.rangeBinIndexPhase, 
			self.vs.maxVal,
			self.vs.processingCyclesOut,
			self.vs.processingCyclesOut1,
			self.vs.rangeBinStartIndex,
			self.vs.rangeBinEndIndex,
			self.vs.unwrapPhasePeak_mm,
			self.vs.outputFilterBreathOut,
			self.vs.outputFilterHeartOut,
			self.vs.heartRateEst_FFT,
			self.vs.heartRateEst_FFT_4Hz,
			self.vs.heartRateEst_xCorr,
			self.vs.heartRateEst_peakCount,
			self.vs.breathingRateEst_FFT,
			self.vs.breathingEst_xCorr,
			self.vs.breathingEst_peakCount,
			self.vs.confidenceMetricBreathOut,
			self.vs.confidenceMetricBreathOut_xCorr,
			self.vs.confidenceMetricHeartOut,
			self.vs.confidenceMetricHeartOut_4Hz,
			self.vs.confidenceMetricHeartOut_xCorr,
			self.vs.sumEnergyBreathWfm,
			self.vs.sumEnergyHeartWfm,
			self.vs.motionDetectedFlag,
			self.vs.rsv[0],self.vs.rsv[1],self.vs.rsv[2],
			self.vs.rsv[3],self.vs.rsv[4],self.vs.rsv[5],
			self.vs.rsv[6],self.vs.rsv[7],self.vs.rsv[8],
			self.vs.rsv[9]) = struct.unpack('2Hf2H2H28f', buf)
		except:
			print("vsos error found:")
			return (False,self.vs)

		return (True,self.vs)
		
		

					
		
	def tlvRead(self,disp):
		#print("---tlvRead---")
		vsdata = vsos
		idx = 0
		lstate = 'idle'
		sbuf = b""
		
		#*******Message TLV Header 1 *******
		mmWaveType1 = 0
		length1 = 0
		#********Message TLV Header 2******
		mmWaveType2 = 0
		length2 = 0
	
		while True:
			ch = self.port.read()
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[idx]:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					#GPIO.output(21, True)	
					idx += 1
					if idx == 8:			
						idx = 0
						lstate = 'header'
						rangeProfile = b""
						sbuf = b""		
						#print("--vital header")
				else:
					idx = 0
					rangeProfile = b""
					return (False,vsdata, rangeProfile)
		
			elif lstate == 'header':
				sbuf += ch
				idx += 1
				if idx == 40:	
					#print("------header-----")
					#print(":".join("{:02x}".format(c) for c in sbuf))
					#print("------header end -----")  
					# [header - Magicword] + [Message TLV header] 
					try:   		  
						'''(self.version , self.totalPackLen, self.platform ,
						self.frameNumber, self.timeCpuCycles, self.numDetectedObj,
						self.numTLVs, rsv, mmWaveType1, length1) = struct.unpack('10I', sbuf)
						'''
						
						(self.hdr.version , self.hdr.totalPackLen, self.hdr.platform ,
						self.hdr.frameNumber, self.hdr.timeCpuCycles, self.hdr.numDetectedObj, 
						self.hdr.numTLVs, self.hdr.rsv, mmWaveType1, length1) = struct.unpack('10I', sbuf) 
		
					
					except:
						print("Improper TLV structure found: ")
						return (False,vsdata, rangeProfile)
					  
					if disp == True:
						self.showHeader()
							#-----[Message TLV header] = 8 bytes ----
						print("mmWave Type 1:\t%d "%(mmWaveType1))
						print("Data 1 Len:\t%d "%(length1)) 
				    
					sbuf = b""
					idx = 0
					lstate = 'vsos'
					#if length1 > 288: 
					length1 = 128 
					#print("header=>data1=>length:"%(length1))
					  
				elif idx > 40:
					idx = 0
					lstate = 'idle'
					return (False, vsdata ,rangeProfile)
				  
								
			elif lstate == 'vsos': #128 Vital Signs Output Status
				sbuf += ch
				idx += 1
				if idx == length1: #128
					#print("-----vital Signs Output Stats----") 
					#print(":".join("{:02x}".format(c) for c in sbuf))
					vflag, vsdata = self.vitalSignsOutputStats(sbuf)
					#print("-----vital Signs Output Stats end ---")
					if not vflag:
						lstate = 'idle'
						return (False,vsdata, rangeProfile)

					idx = 0 
					lstate = 'mTLVh'
					sbuf = b""
				
				elif idx > length1:
					idx = 0
					lstate = 'idle'
					return (False,vsdata, rangeProfile)
						 
			elif lstate == 'mTLVh':
				sbuf += ch
				idx += 1
				if idx == 8:
					#print(":".join("{:02x}".format(c) for c in sbuf))
					idx = 0  
					try:   		  
						mmWaveType2 , length2 = struct.unpack('2I',sbuf) 
					except:
						print("Improper Type2, length structure found: ")
						return (False,vsdata, rangeProfile)
						
					#print("mmWave Type 2:\t%d "%(mmWaveType2))
					#print("Data 2 Len:\t%d "%(length2))
					#print("mTLVh---state")  
					if length2 > 252:
						length2 = 252
					lstate = 'rangeProfile'
					rangeProfile = b""
					#[Message TLV header] = 8 bytes
					
				   
				elif idx > 8:
					idx = 0
					lstste = 'idle'
				  
			elif lstate == 'rangeProfile':
				rangeProfile += ch
				idx += 1
				if idx == length2:
					idx = 0
					try:
						fmt = '{:d}h'.format(int(length2/2))
						xd = struct.unpack(fmt,rangeProfile)
					except:
						print("Improper rangeProfile data found: ")
						return (False,vsdata, rangeProfile)
        
					#print("---------rangeProfile:" + str(len(rangeProfile)))
					#print(":".join("{:02x}".format(c) for c in rangeProfile))
					#GPIO.output(21, False)
					#print(xd)
					return (True,vsdata, list(xd))
				elif idx > length2:
					idx = 0
					lstate = 'idle'





# Medium Range Radar Raw Data (MRR)
# ver:0.0.1
# 2021/12/31
# parsing Medium Range Radar
# hardware:(Batman-601):  
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: V1,V2,V3,v4 Raw data
# v0.0.1 : 2021/12/31 release
# v0.0.2 : 2022/02/07 format revised
# v0.1.0 : 2022/06/01 tune to fit v3
# v1.0.0 : 2022/06/08 change get Uart method
# v1.0.1 : 2022/06/09 added headerShowAll()
 

import serial
import time
import struct
import pandas as pd
import numpy as np

class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	cpuProcessTime = 0
	numOBJs = 0
	numTLVs = 0
	subFrameNumber = 0

class byteInfo:
	headerByte = 32


class MRR:
    
	magicWord =  [b'\x01',b'\x02',b'\x03',b'\x04',b'\x05',b'\x06',b'\x07',b'\x08',b'\0x99']
	magicLen = 8
	port = ''
	hdr = header
	hdr0 = header()
	hdr1 = header()
	info = byteInfo
	hdrA = [hdr0,hdr1]
	headerDic = {0:hdr0, 1:hdr1}
	
	clearQBuf = True
	# provide csv file dataframe
	# real-time 
	v1_col_names_rt = ['fN','type','doppler','peakVal','X' ,'Y','Z']  
	v2_col_names_rt = ['fN','type','ccX','ccY','csX','csY']
	v3_col_names_rt = ['fN','type', 'tX','tY','velX','velY','csX','csY']
	v4_col_names_rt = ['fN','type','parkingA']  
	
	# read from file for trace point clouds
	fileName = ''
	v1_col_names = ['time','fN','type','doppler','peakVal','X' ,'Y','Z']
	v2_col_names = ['time','fN','type','ccX','ccY','csX','csY']
	v3_col_names = ['time','fN','type', 'tX','tY','velX','velY','csX','csY']
	v4_col_names = ['time','fN','parkingA']  
	sim_startFN = 0
	sim_stopFN  = 0 
	
	v1simo = []
	v2simo = []
	v3simo = []
	v4simo = []
	
	buf = b''
	
	v1 = ([])
	v2 = ([])
	v3 = ([])
	v4 = ([])
	v1df = ([])
	v2df = ([])
	v3df = ([])
	v4df = ([])
	
	def clearBuffer_v124(self):
		if self.clearQBuf == True:
			self.v1 = ([])
			self.v2 = ([])
			#self.v3 = ([])
			self.v4 = ([])
			self.v1df = ([])
			self.v2df = ([])
			#self.v3df = ([])
			self.v4df = ([])
			
	def clearBuffer_v3(self):
		if self.clearQBuf == True:
			self.v3 = ([])
			self.v3df = ([])

	# add for interal use
	tlvLength = 0

	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm = True #Observed StateMachine: True Show message
	 
	
	def __init__(self,port):
		self.port = port
		print("(jb)Medium Range Radar(MRR) raw Data lib initial")
		print("(jb)For Hardware:Batman-601(ISK)")
		print("(jb)Version: v0.1.0")
		print("(jb)Hardware: IWR-6843 ES2.0")
		print("(jb)Firmware: MRR")
		print("(jb)UART Baud Rate:921600")
		print("==============Info=================")
		print("Output: V1,V2,V3,V4 data:(RAW)")
		print("V1: Detected Object")
		print("V1 structure: [(hdr,Doppler,peakVal,X,Y,Z),......]")
		print("V2: Cluster")
		print("V2 structure: [(hdr,ClusterX,ClusterY,ClusterSizeX,ClusterSizeY),....]")
		print("V3: Tracking Object")
		print("V3 structure: [(hdr,TrackingX,TrackingY,velX,velY,ClusterSizeX,ClusterSizeY)....]")
		print("V4: Parking Assist")
		print("V4 [hdr,....] length:32")
		print("===================================")
	 
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	def headerShow(self):
		print("***Header***********")  # 32 bytes
		print("Version:     \t%x "%(self.hdr.version))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("Inter-frame Processing Time:\t{:d} us".format(self.hdr.cpuProcessTime))
		print("number of Objects:     \t%d "%(self.hdr.numOBJs))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("subframe#  : \t%d "%(self.hdr.subFrameNumber))
		print("***End Of Header***") 
		
	
	def headerShowAll(self):
		for i in range(2):
			print(f'hdrA({i})  : fn = { self.headerDic[i].frameNumber} ')
			print("***Header***********")  # 32 bytes
			print("Version:     \t%x "%( self.headerDic[i].version))
			print("TotalPackLen:\t%d "%( self.headerDic[i].totalPackLen))
			print("Platform:    \t%X "%( self.headerDic[i].platform))
			print("PID(frame#): \t%d "%( self.headerDic[i].frameNumber))
			print("Inter-frame Processing Time:\t{:d} us".format( self.headerDic[i].cpuProcessTime))
			print("number of Objects:     \t%d "%( self.headerDic[i].numOBJs))
			print("numTLVs:     \t%d "%( self.headerDic[i].numTLVs))
			print("subframe#  : \t%d "%( self.headerDic[i].subFrameNumber))
			print("***End Of Header***") 
	
	def list2df(self,chk,l1,l2,l3,l4):
		#print("---------------list2df: v1----------------")
		#print(l1)
		
		ll1 = pd.DataFrame(l1,columns=self.v1_col_names_rt)
		ll2 = pd.DataFrame(l2,columns=self.v2_col_names_rt)
		ll3 = pd.DataFrame(l3,columns=self.v3_col_names_rt)
		ll4 = pd.DataFrame(l4,columns=self.v4_col_names_rt)
		#print("------ll1---------list2df: v1----------------")
		#print(ll1)
		return (chk,ll1,ll2,ll3,ll4)
		
	 
	#################################################################################
	# FUNCTION: jb_getUartBuf(); get from serial port into buffer uartBuf[]
	#################################################################################
	JB_syncFlag = 0
	def jb_getUartBuf(self,disp = None):
		idx = 0
		self.buf = b''
		self.buf1 = b''
		bufDic = {}
		version = 0 
		totalPackLen = 0 
		platform = 0
		frameNumber = 0 
		cpuProcessTime = 0
		numOBJs = 0 
		numTLVs = 0
		subFrameNumber = 0
		hdr0 = header()
		hdr1 = header()
		self.headerDic = {0:hdr0, 1:hdr1}
		
		while True:
			if self.JB_syncFlag == 0:
				ch = self.port.read() # here ch type is byte
				if ch == self.magicWord[idx]: # 0..7
					self.buf += ch
					idx += 1 # 1..8
					if idx == 8:
						idx = 8
						self.JB_syncFlag = 1
						self.JB_t0 = time.time()
						self.JB_tMagic = time.time()
			else:
				self.JB_syncFlag = 0 # magic word ch mismatch loop back again
				idx = 0 # init
				self.buf = b""
				
			if self.JB_syncFlag == 1:
				# read next 32 bytes for header version, len, platform, fn
				for i in range(self.info.headerByte):
					ch = self.port.read()
					self.buf += ch
				if len(self.buf) == (self.magicLen + self.info.headerByte): 
					try: 
						(version,totalPackLen,platform,frameNumber,cpuProcessTime,numOBJs,numTLVs,subFrameNumber
						) = struct.unpack('8I', self.buf[8:40])
						
						(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,self.hdr.frameNumber,
						self.hdr.cpuProcessTime,self.hdr.numOBJs,self.hdr.numTLVs,self.hdr.subFrameNumber
						) = (version,totalPackLen,platform,frameNumber,cpuProcessTime,numOBJs,numTLVs,subFrameNumber)
						#if self.dbg == True:
						if disp:
							self.headerShow()
						 
					except:
						#if self.dbg == True:
						print("(Header)Improper TLV structure found: ")
						return (bufDic,self.headerDic)
					
					fn = self.hdr.frameNumber
					self.JB_syncFlag = 2
					
				else:
					self.JB_syncFlag = 0 # len mismatch loop back again
					idx = 0
					self.buf = b""

			if self.JB_syncFlag == 2:
				
				dataLen =  totalPackLen - self.magicLen - self.info.headerByte
				#print(f'----------JB_sync=2  data length: {dataLen}------------')
				r = self.port.read(dataLen) # read rest of bytes
				self.buf += r
				#self.port.flushInput() 
				dt = time.time()- self.JB_t0
				self.JB_tBuf = time.time()
				
				if self.hdr.subFrameNumber == 1:
					bufDic[1]    = self.buf
					(hdr1.version,hdr1.totalPackLen,hdr1.platform,hdr1.frameNumber,hdr1.cpuProcessTime,hdr1.numOBJs,hdr1.numTLVs,hdr1.subFrameNumber
					) = (version,totalPackLen,platform,frameNumber,cpuProcessTime,numOBJs,numTLVs,subFrameNumber)
					#headerDic[1] = hdr1
					self.JB_syncFlag == 0
					#print(f"--------1-----------{hdr1.totalPackLen}--------{ time.time() - self.JB_t0 }")
					self.port.flushInput() 
					return (bufDic,self.headerDic)
				 
				if self.hdr.subFrameNumber == 0:
					bufDic[0] = self.buf
					
					(hdr0.version,hdr0.totalPackLen,hdr0.platform,hdr0.frameNumber,hdr0.cpuProcessTime,hdr0.numOBJs,hdr0.numTLVs,hdr0.subFrameNumber
					) =  (version,totalPackLen,platform,frameNumber,cpuProcessTime,numOBJs,numTLVs,subFrameNumber)
					#headerDic[0] = hdr0
					self.JB_syncFlag == 0
					idx = 0
					self.buf = b""
					#print(f"--------0-----------{hdr0.totalPackLen}--------{ time.time() - self.JB_t0 }")
				 
	#################################################################
#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
# output:(return parameter)
# (pass_fail, v1, v2, v3, v4)
#  pass_fail: True: Data available    False: Data not available
#
#  v1: Detected Object List infomation
#  v2: Cluster Output information
#  v3: Tracking Output information
#  v4: Parking Assist Infomation
#
#  chk:  
#   0: empty: tlv = 0
#   1: data output
#  
#   10: idle
#   99: error
#
#
	def tlvRead(self,disp,df = None):
		#print("---tlvRead---")
		#ds = dos
		typeList   = [1,2,3,4]
		typeString = ['HDR','V1','V2','V3','V4']
		msgString  = {'HDR':"",'V1':'V1 Detected Object','V2':'V2 Cluster','V3':'V3 Tracking Object','V4':'V4 Parking Assist'}
		idx = 0
		lstate = 'idle'
		sbuf = b""
		tlvCount = 0
		pbyte = 16
		
		typeCnt = 0 # for check with tlvCount, if typeCnt equal to tlvCcount the process will completed 
		idxFrom = 0
		idxTo = 0
		magicLen  = 8
		headerLen = 32
		typeLen   = 8
		objUnitLen = 4
		
		''' 
		(uartBuf,hdrA) = self.jb_getUartBuf()
		print(f'uartBuf[0]:\n {uartBuf[0]}\n--------------------\nuartBuf[1]:\n {uartBuf[0]}\n')
		'''
		
		hdr_sf = 0 
		while True:
			
			if lstate == 'idle':
				if self.sm == True:
					print(f'(SM) {lstate}') 
					
				(uartBuf,hdrA) = self.jb_getUartBuf(disp = disp)
				lstate = 'block0'
				idxFrom = 0
				idxTo = 0
				
			if lstate == 'block0': #subFrameNumber = 0 
				hdr_sf = 0
				if self.sm == True:
						print(f'(SM) {lstate}') 
						
				if hdrA[0].numOBJs == 0 and hdrA[0].numTLVs == 0 and hdrA[0].subFrameNumber == 0:
					self.v3 = []; self.v3df = []
					lstate = 'block1'
				else:
					try:
						sbuf = uartBuf[0][40:]
					except:
						return self.list2df(99,self.v1df,self.v2df,self.v3df,self.v4df) if (df == 'DataFrame') else (99,self.v1,self.v2,self.v3,self.v4)
					#print("".join("{:02x}".format(c) for c in sbuf)) 
					idxFrom = 0
					idxTo = typeLen + objUnitLen
					tlvCount = hdrA[0].numTLVs
					lstate = 'Units'
			
			if lstate ==  'block1':
				hdr_sf = 1
				sbuf = b''
				sbuf = uartBuf[1][40:] #remove header
				if self.sm == True:
						print(f'(SM) {lstate}') 
						
				'''
				print(f'uartBuf-header Length= {len(sbuf)} objs:{hdrA[1].numOBJs} tlv#:{hdrA[1].numTLVs} subFrame:{hdrA[1].subFrameNumber}')
				print("".join("{:02x}".format(c) for c in sbuf))  
				'''
				if hdrA[1].numOBJs == 0 and hdrA[1].numTLVs == 0 and hdrA[1].subFrameNumber == 1:
					self.v1 = [];self.v1df = []; self.v2 = [];self.v2df = [];self.v4 = [];self.v4df = []
					return self.list2df(10,self.v1df,self.v2df,self.v3df,self.v4df) if (df == 'DataFrame') else (10,self.v1,self.v2,self.v3,self.v4)
				idxFrom = 0
				idxTo = typeLen + objUnitLen
				tlvCount = hdrA[1].numTLVs
				lstate = 'Units'
		
			if lstate == 'Units':
				#print(f"============= Units ================== from:to= [{idxFrom}:{idxTo}]")
				#print("".join("{:02x}".format(c) for c in sbuf[idxFrom:idxTo])) 
				ttype,self.tlvLength, objs,xyzQFormat = struct.unpack('2I2H', sbuf[idxFrom:idxTo]) 
				if ttype not in typeList:
					return self.list2df(99,self.v1df,self.v2df,self.v3df,self.v4df) if (df == 'DataFrame') else (99,self.v1,self.v2,self.v3,self.v4)
					
				vUnit = 1 / 2**xyzQFormat
				if disp:
					print(f'type/length: v{ttype}/{self.tlvLength} objs:{objs} xyzQFormat:{xyzQFormat} unit:{vUnit}')
				lstate = typeString[ttype]
				
				idxFrom = idxTo
				idxTo = idxFrom + self.tlvLength - objUnitLen 
				#print(f"============= Units -> {lstate} ====== from:to= [{idxFrom}:{idxTo}]")
			
			if lstate == 'V1' or lstate == 'V2' or lstate == 'V4' or lstate == 'V3':
				if self.sm == True:
						print(f'(SM) {lstate}') 
				tlvCount -= 1
				self.typeFunc(typeT = lstate ,vUnit= vUnit ,objs=objs,ldata = sbuf[idxFrom:idxTo], df = df)
				idxFrom = idxTo
				idxTo = idxFrom + typeLen + objUnitLen 
				lstate = 'Units'
			
			if tlvCount == 0 and hdr_sf == 0:
				lstate = 'block1'
				idxTo = 0
				idxFrom = 0
				
			if tlvCount == 0 and hdr_sf == 1:
				lstate = 'idle'
				return self.list2df(1,self.v1df,self.v2df,self.v3df,self.v4df) if (df == 'DataFrame') else (1,self.v1,self.v2,self.v3,self.v4)
			
			
	def typeFunc(self,typeT = None ,vUnit= None,objs= None,ldata = None, df = None):
		#print('typeFunc===> {}  vUnit:{:}'.format(typeT,vUnit))
		#print(":".join("{:02x}".format(c) for c in ldata))
		
		if typeT == 'V1':
			self.v1df = ([])
			self.v1   = ([])
			for i in range(objs):
				start = i * 10
				end   = (i+1) * 10
				(dt,pt,xt,yt,zt) = struct.unpack('hH3h', ldata[start:end]) # ori 5h
				d = dt * vUnit
				p = pt * vUnit
				x = xt * vUnit
				y = yt * vUnit
				z = zt * vUnit
				if self.dbg == True:
					print("V1 =>  d:{:.4f} p: {:.4f} x:{:.4f} y:{:.4f} z:{:.4f}".format(d,p,x,y,z))
				#v1_col_names_rt = ['fN','type','doppler','peakVal','X','Y','Z']
				if (df == 'DataFrame'):
					self.v1df.append((self.hdr.frameNumber,'v1',d,p,x,y,z))
					
				else:
					self.v1.append((self.hdr.frameNumber,d,p,x,y,z))
				
		if typeT == 'V2':
			self.v2df = ([])
			self.v2   = ([])
			for i in range(objs):
				start = i * 8
				end   = (i+1) * 8
				(ccXt,ccYt,csXt,csYt) = struct.unpack('4h', ldata[start:end])  # ori 4h
				ccX  =  vUnit * ccXt
				ccY  =  vUnit * ccYt
				csX  =  vUnit * csXt
				csY  =  vUnit * csYt
				if self.dbg == True:
					print("v2 => X:{:.4f} Y:{:.4f}  XD:{:.4f}  YD:{:.4f}".format(ccX,ccY,csX,csY))
				if (df == 'DataFrame'):
					self.v2df.append((self.hdr.frameNumber,'v2',ccX,ccY,csX,csY))
				else:
					self.v2.append((self.hdr.frameNumber,ccX,ccY,csX,csY)) 
					
		if typeT == 'V3': 
			self.v3 = ([])
			self.v3df = ([])
			for i in range(objs):
				start = i * 12
				end   = (i+1) * 12
				(Xt,Yt,XDt,YDt,Xsizet,Ysizet) = struct.unpack('6h', ldata[start:end]) # ori 6h
				X 	  =  vUnit * Xt
				Y     =  vUnit * Yt
				XD    =  vUnit * XDt
				YD    =  vUnit * YDt
				Xsize =  vUnit * Xsizet
				Ysize =  vUnit * Ysizet
				if self.dbg == True:
					print("v3 => X:{:.4f} Y:{:.4f}  XD:{:.4f}  YD:{:.4f} Xsize:{:.4f} Ysize:{:.4f}".format(X,Y,XD,YD,Xsize,Ysize))
				if (df == 'DataFrame' ):
					self.v3df.append((self.hdr.frameNumber,'v3',X,Y,XD,YD,Xsize,Ysize))
				else: 
					self.v3.append((self.hdr.frameNumber,X,Y,XD,YD,Xsize,Ysize))
					
		if typeT == 'V4':
			self.v4df = ([])
			self.v4   = ([])
			for i in range(objs):
				astA = [] 
				for i in range(objs):
					start = i * 2
					end   = (i+1) * 2
					ast, = struct.unpack('H', ldata[start:end]) 
					astA.append(ast * vUnit)
					
			if (df == 'DataFrame'):
				self.v4df.append((self.hdr.frameNumber,'v4',astA))
			else:
				self.v4.append((self.hdr.frameNumber,astA))
	
	

	def getRecordData(self,frameNum):
		s_fn = frameNum + self.sim_startFN
		#print("frame number:{:}".format(s_fn))
		v1d = self.v1simo[self.v1simo['fN'] == s_fn]
		
		v2d = self.v2simo[self.v2simo['fN'] == s_fn]
		v3d = self.v3simo[self.v3simo['fN'] == s_fn]
		v4d = self.v4simo[self.v4simo['fN'] == s_fn]
		return (v6d,v7d,v8d,v9d)
		
		
	'''
	def readFile(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		#          ['time','fN','type','elv','azimuth','range' ,'doppler','sx', 'sy', 'sz']
		df = pd.read_csv(self.fileName, names = self.v6_col_names, skiprows = [0,10,11,12],dtype={'fN': int,'elv': float,'azimuth':float,'range':float,'doppler':float,'sx':float,'sy':float,'sz':float}) 
		df.dropna()
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		print(df)
		print(df.info())
		#print(df.info(memory_usage="deep")) 
		
		v1simOri = df[(df.type == 'v1')]
		#print("-------------------v1simo------------:{:}".format(v6simOri.shape))
									 
		self.v1simo = v1simOri.loc[:,['fN','type','range','azimuth','elv' ,'doppler','sx', 'sy', 'sz']]
		
		if len(self.v1simo):
			self.sim_startFN = df['fN'].values[0]
			self.sim_stopFN  = df['fN'].values[-1]
		#print(self.v6simo)
		
		self.v6simo['elv'] = self.v6simo['elv'].astype(float, errors = 'raise') 
		
		df7 = pd.read_csv(self.fileName, names = self.v7_col_names, skiprows = [0])  
		
		v7simc = df7[df7['type'] == 'v7']
		self.v7simo  = v7simc.loc[:,['fN','posX','posY','velX','velY','accX','accY','posZ','velZ','accZ','tid']]
		
		v8simc = df[df['type'] == 'v8']
		self.v8simo  = v8simc.loc[:,['fN','elv']]
		#print(self.v8simo)
		
		v9simc = df[df['type'] == 'v9']
		self.v9simo  = v9simc.loc[:,['fN','type','range','azimuth']]
		self.v9simo.columns = ['fN','type','snr','noise']
	
		return (self.v6simo,self.v7simo,self.v8simo,self.v9simo)
	'''





#
# Short Range Radar (SRR)
# ver:0.0.1
# date: 2019/10/03 release
# 
import numpy as np
import struct
import serial

#Frame Header
class header:
	version = "0.0"
	totalPackLen =0
	platform = ""
	frameNumber = 0
	timeCpuCycles = 0
	numDetectedObj = 0
	numTLVs = 0
	tvlHeaderLen = 8
	subFrameNumber = 0


class SRR:
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	port  = ""
	hdr = header

	version = "0.0"
	platform = "iwr1642" 
	sm = False
	checkPort = False
	
	def __init__(self,port):
		self.port = port
		print("*******************************************")
		print("***Short Range Radar init")
		print("***Check TLV state: Call stateMachine(True)")
		print("***Check header: Call getHeader()")
		print("***print info:  tlvRead(True)")
		print("*** More information check Github")
		print("*******************************************")
	
	def srr_test(self):
		print("*********SRR Called**********")
	
	def getHeader(self):
		return self.hdr
		
	def stateMachine(self,ft):
		self.sm = ft
	
	def checkUARTPort(self,ft):
		self.checkPort = ft
	
	def headerShow(self):
		print("***header***********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("Platform:    \t%X "%(self.hdr.platform))
		#print("Time Stamp:  \t%d "%(self.hdr.timestamp))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("timeCpuCycles#  : \t%d "%(self.hdr.timeCpuCycles))
		print("numDetectedObjs:     \t%d "%(self.hdr.numDetectedObj))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("subframe#  : \t%d "%(self.hdr.subFrameNumber))
		#print("Check Sum   :\t{:x}".format(self.hdr.checksum))
		 
		
#TLV: Type-Length-Value
#read TLV data
	
	def tlvRead(self,disp):
		typeDesc = ["idel","V1","V2","V3","V4"]
		#print("---tlvRead---")
		idx = 0
		#version = ""
		lstate = 'idle'
		sbuf = b""
		#tlvLength = 0
		numTLVsCnt = 0

		#*******Message TLV Header 1 *******
		tlvNDO = 0
		tlvNDOCnt = 0
		xyzQFormat = 0
		devParaQ = 0
		 
		#V1
		jb_doppler_v1 = 0
		jb_peakValue = 0
		jb_range_v1 = 0
		jb_x = 0
		jb_y = 0
		#V2
		jb_xc = 0.0
		jb_yc = 0.0
		jb_xcSize = 0.0
		jb_ycSize = 0.0
		#V3
		jb_xt = 0.0  #for Range
		jb_yt = 0.0  #for Range
		jb_vxt = 0.0 #for Doppler
		jb_vyt = 0.0 #for Doppler
		jb_xtSize = 0.0
		jb_ytSize = 0.0
		jb_tRange = 0.0    #sqrt(jb_xt^2+jb_yt^2)
		jb_tDoppler = 0.0  #
		
		#V4
		jb_val = 0.0
		
		v1 = () #objectDet
		v2 = () #ClusterData
		v3 = () #TRACK Data
		v4 = () #PARK Data"
		v6 = () #statsInfo
	
		while True:
			ch = self.port.read()
			if self.checkPort:
				print(str(ch))
				
			if lstate == 'idle':
			#print("state:" + lstate)
				if ch == self.magicWord[idx]:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)))
					#GPIO.output(21, True)
					idx += 1
					if idx == 8:
						idx = 0
						lstate = 'header'
						sbuf = b""
				else:
					idx = 0
					return (0,v1,v2,v3,v4)
		
			elif lstate == 'header': #header
				sbuf += ch
				idx += 1
				#print(str(ch))
				if idx == 32: #32 + 12 = 44
					# print(":".join("{:02x}".format(c) for c in sbuf))
					#print("------header end -----")
					# [header - Magicword] + [Message TLV header] 
					try:   		  
						(self.hdr.version , self.hdr.totalPackLen, self.hdr.platform , self.hdr.frameNumber, self.hdr.timeCpuCycles,
						 self.hdr.numDetectedObj, self.hdr.numTLVs, self.hdr.subFrameNumber) = struct.unpack('8I', sbuf) 
						
					except:
						print("Improper TLV structure found: ")
						return (0,v1,v2,v3,v4)
						break
					
					if disp == True:
						self.headerShow()
					
					numTLVsCnt = self.hdr.numTLVs 
					if self.hdr.numTLVs == 0:
						lstate = 'idle'
						return (0,v1,v2,v3,v4)
					
					sbuf = b""
					idx = 0
					v1=[]
					v2=[]
					v3=[]
					v4=[]
					lstate = 'TL'
					if self.sm:
						print("------------(header)-numTLVs:{:d} ---subFrame:{:d} --->(TL)".format(self.hdr.numTLVs ,self.hdr.subFrameNumber))
					
				elif idx > 44: # not use
					idx = 0
					lstate = 'idle'
			
			elif lstate == 'TL':
				sbuf += ch
				idx += 1
				if idx == 12: #
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						ttype,self.tlvLength , tlvNDO ,xyzQFormat = struct.unpack('2Ihh',sbuf)
						if self.sm:
							print("(TL) type: {:d} length: {:d} numDetectedObj: {:d}  Qformat:{:d} ".format(ttype,self.tlvLength,tlvNDO,xyzQFormat))
						devParaQ = 2**xyzQFormat
						
					except:
						print("Improper TLV:TL structure found:")
						return (0,v1,v2,v3,v4)
					if ttype < 5 :
						lstate = typeDesc[ttype]

					if ttype == 6:
						lstate = 'V6'
						
					if ttype > 7:
						return (0,v1,v2,v3,v4)
					
					if self.sm:	
						print("From (TL) ==> ({:})".format(lstate))
						
					numTLVsCnt -= 1
					tlvNDOCnt = 0
					sbuf = b""
					idx = 0
					
			elif lstate == "V1": #Object Detected
				sbuf += ch
				idx += 1
				if idx == 8:
					#print("(V1)numTLV:{:d} numTLVsCnt:{:d} tlvNDO:{:d}  getObject:{:d}".format(self.hdr.numTLVs,numTLVsCnt,tlvNDO,tlvNDOCnt))
					try:
						jb_doppler_v1,jb_peakValue,jb_x,jb_y = struct.unpack('hHhh' , sbuf)
						jb_x = float(jb_x)/devParaQ
						jb_y = float(jb_y)/devParaQ
						jb_doppler_v1 = float(jb_doppler_v1)/devParaQ
						jb_range_v1 = np.sqrt(jb_x * jb_x + jb_y * jb_y)
						jb_peakValue = float(jb_peakValue)/devParaQ
						v1.append((tlvNDOCnt,jb_x,jb_y,jb_doppler_v1,jb_range_v1,jb_peakValue))
						
					except:
						print("Improper TLV:V1 structure found:")
						return (0,v1,v2,v3,v4)
						
					tlvNDOCnt += 1
					sbuf = b''
					idx = 0
					 
					if tlvNDOCnt >= tlvNDO:
						if self.sm:
							print("(V1) completed ==> (TL)")
						lstate = 'TL'
						if numTLVsCnt == 0:
							lstate = 'idle'
							return (1,v1,v2,v3,v4)
					else:
						lstate = 'V1'
					
			elif lstate == "V2": #Cluster
				sbuf += ch
				idx += 1
				if idx == 8:
					#print("(V2)numTLV:{:d} numTLVsCnt:{:d} tlvNDO:{:d}  getObject:{:d}".format(self.hdr.numTLVs,numTLVsCnt,tlvNDO,tlvNDOCnt))
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						jb_xc,jb_yc,jb_xcSize,jb_ycSize = struct.unpack('4h' , sbuf)
						v2.append((tlvNDOCnt,float(jb_xc)/devParaQ,float(jb_yc)/devParaQ,float(jb_xcSize)/devParaQ,float(jb_ycSize)/devParaQ))
						#print("(V2)x:{:f} y:{:f} xSize:{:f} ySize:{:f}".format(float(jb_xc)/devParaQ,float(jb_yc)/devParaQ,float(jb_xcSize)/devParaQ,float(jb_ycSize)/devParaQ))
					except:
						print("Improper TLV structure found:")
						return (0,v1,v2,v3,v4)
						
					tlvNDOCnt += 1
					sbuf = b''
					idx = 0
					 
					if tlvNDOCnt >= tlvNDO:
						if self.sm:
							print("(V2) completed ==> (TL)")
						lstate = 'TL'
						if numTLVsCnt == 0:
							lstate = 'idle'
							return (1,v1,v2,v3,v4)
					else:
						lstate = 'V2'
					
			elif lstate == "V3": #Track
				sbuf += ch
				idx += 1
				if idx == 12:
					#print("(V3)numTLV:{:d} numTLVsCnt:{:d} tlvNDO:{:d}  getObject:{:d}".format(self.hdr.numTLVs,numTLVsCnt,tlvNDO,tlvNDOCnt))
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						x0,y0,vx0,vy0,xs0,ys0 = struct.unpack('6h' , sbuf)
						jb_xt = float(x0)/devParaQ
						jb_yt = float(y0)/devParaQ
						jb_vxt = float(vx0)/devParaQ
						jb_vyt = float(vy0)/devParaQ
						jb_xtSize =  float(xs0)/devParaQ
						jb_ytSize =  float(ys0)/devParaQ
						jb_tRange = np.sqrt(jb_xt * jb_xt + jb_yt * jb_yt)
						jb_tDoppler = (jb_xt * jb_vxt + jb_yt * jb_vyt) / jb_tRange
						v3.append((tlvNDOCnt,jb_xt,jb_yt,jb_xtSize,jb_ytSize,jb_vxt,jb_vyt,jb_tRange,jb_tDoppler))
						#print("devParaQ:{:f}".format(devParaQ))
						#print("(V3)x:{:f} y:{:f} xSize:{:f} ySize:{:f} range:{:f} dpler:{:f}".format(jb_xt,jb_yt,jb_xtSize,jb_ytSize,jb_tRange,jb_tDoppler))
					except:
						print("Improper TLV:V3 structure found or Operation Error:")
						return (0,v1,v2,v3,v4)
						
					tlvNDOCnt += 1
					sbuf = b''
					idx = 0
					 
					if tlvNDOCnt >= tlvNDO:
						if self.sm:
							print("(V3) completed ==> (TL)")
						#print("----V3---")
						#print(v3)
						lstate = 'TL'
						if numTLVsCnt == 0:
							lstate = 'idle'
							return (1,v1,v2,v3,v4)
					else:
						lstate = 'V3'
				
			elif lstate == "V4": #Parking_assist
				sbuf += ch
				idx += 1
				if idx == 2:
					#print("(V4)numTLV:{:d} numTLVsCnt:{:d} tlvNDO:{:d}  getObject:{:d}".format(self.hdr.numTLVs,numTLVsCnt,tlvNDO,tlvNDOCnt))
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						pv, = struct.unpack('h' , sbuf)
						jb_val = float(pv)/devParaQ
						v4.append(jb_val)
					except:
						print("Improper TLV:V4 structure found or Operation Error:")
						return (0,v1,v2,v3,v4)
						
					tlvNDOCnt += 1
					sbuf = b''
					idx = 0
					 
					if tlvNDOCnt >= tlvNDO:
						if self.sm:
							print("(V4) completed ==> (TL)")
						#print("----V4---")
						#print(v4)
						lstate = 'TL'
						if numTLVsCnt == 0:
							lstate = 'idle'
							
							
							
							
							
							return (1,v1,v2,v3,v4)
					else:
						lstate = 'V4'
			'''			
			elif lstate == "V6":
				print("--------V6---------")
				lstate = 'idle'
				return (1,v1,v2,v3,v4)
			'''	
			
	 




# 3D People Counting-OSD 3D Raw Data (pc3d)
# ver:0.0.1
# 2019/10/16
# parsing 3D People Overhead Counting data 
# hardware:(Batman-301)ODS IWR6843 ES2.0 only
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: V6,V7,V8 Raw data
# v0.0.1 : 2019/10/26 release
# v0.0.2 : 2019/10/27 release



import serial
import time
import struct
#import numpy as np

class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	subframeNumber = 0
	chirpMargin = 0
	frameMargin = 0 
	trackProcessTime = 0
	uartSendTime = 0
	numTLVs = 0
	checksum = 0

class unitS:
	elevationUnit:float = 0.0
	azimuthUnit : float = 0.0
	dopplerUnit :float = 0.0
	rangeUnit :float = 0.0
	snrUnit :float = 0.0

class Pc3d:
	
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	port = ""
	hdr = header
	u = unitS
	
	# add for PMB interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm = False #Observed StateMachine: True Show message
	plen = 16 
	
	def __init__(self,port):
		self.port = port
		print("(jb)3D People Overhead Counting initial")
		print("(jb)For Hardware:Batman-301(ODS)")
		print("(jb)Hardware: IWR-6843 ES2.0")
		print("(jb)Firmware: FDS")
		print("(jb)UART Baud Rate:921600")
		print("Output: V6,V7,V8 data:(RAW)")
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	def headerShow(self):
		print("***header***********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("Inter-frame Processing Time:\t{:d} us".format(self.hdr.trackProcessTime))
		print("UART Send Time:\t{:d} us".format(self.hdr.uartSendTime))
		print("Inter-chirp Processing Margin:\t{:d} us".format(self.hdr.chirpMargin))
		print("Inter-frame Processing Margin:\t{:d} us".format(self.hdr.frameMargin))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("Check Sum   :\t{:x}".format(self.hdr.checksum))
		print("***End Of Header***") 
			
	#for class internal use
	def tlvTypeInfo(self,dtype,count,dShow):
		
		sbyte = 8  #tlvHeader Struct = 8 bytes
		unitByte = 20 
		dataByte = 0
		pString = ""
		nString = "numOfPoints :"
		stateString = "V6-unit"
		if dtype == 6:
			unitByte = 20  #pointUnit= 20bytes
			sbyte = 8      #tlvHeader Struct = 8 bytes
			dataByte= 8    #pointStruct 8bytes:(elevation,azimuth,doppler,range,snr)
			pString = "Point Cloud TLV"
			nString = ""
		elif dtype == 7:
			unitByte = 0   #pointUnit= 0bytes 
			sbyte = 8 	   #tlvHeader Struct = 8 bytes
			dataByte = 40  #target struct 40 bytes:(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ)  
			pString = "Target Object TLV"
			nString = "numOfObjects:"
			stateString = "V7"
		elif dtype == 8:
			unitByte = 0 #pointUnit= 0bytes 
			sbyte = 8    #tlvHeader Struct = 8 bytes
			dataByte = 1 #targetID = 1 byte
			pString = "Target Index TLV"
			nString = "numOfIDs"
			stateString = "V8"
		else:
			unitByte = 0
			sbyte = 1
			pString = "*** Type Error ***"
			stateString = 'idle'
		 
		retCnt = count - unitByte -sbyte
		nPoint = retCnt / sbyte
		if dShow == True:
			print("-----[{:}] ----".format(pString))
			print("tlv Type({:2d}B):  \t{:d}".format(sbyte,dtype))
			print("tlv length:      \t{:d}".format(count)) 
			print("{:}      \t{:d}".format(nString,int(nPoint)))
			print("value length:    \t{:d}".format(retCnt))  
		
		return unitByte,stateString, sbyte, dataByte,retCnt, int(nPoint)
		
#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
# output:(return parameter)
# (pass_fail, v6, v7, v8)
#  pass_fail: True: Data available    False: Data not available
#  v6: point cloud infomation
#  v7: Target Object information
#  v8: Target Index information
#
	def tlvRead(self,disp):
		#print("---tlvRead---")
		#ds = dos
		typeList = [6,7,8]
		idx = 0
		lstate = 'idle'
		sbuf = b""
		lenCount = 0
		unitByteCount = 0
		dataBytes = 0
		numOfPoints = 0
		tlvCount = 0
		pbyte = 16
		v6 = ([])
		v7 = ([])
		v8 = ([])
	
		while True:
			try:
				ch = self.port.read()
			except:
				return (False,v6,v7,v8)
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[idx]:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					idx += 1
					if idx == 8:
						idx = 0
						lstate = 'header'
						rangeProfile = b""
						sbuf = b""
				else:
					#print("not: magicWord state:")
					idx = 0
					rangeProfile = b""
					return (False,v6,v7,v8)
		
			elif lstate == 'header':
				sbuf += ch
				idx += 1
				if idx == 40: 
					#print("------header-----")
					#print(":".join("{:02x}".format(c) for c in sbuf)) 	 
					#print("len:{:d}".format(len(sbuf))) 
					# [header - Magicword]
					try: 
						(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,
						self.hdr.frameNumber,self.hdr.subframeNumber,
						self.hdr.chirpMargin,self.hdr.frameMargin,self.hdr.trackProcessTime,self.hdr.uartSendTime,
						self.hdr.numTLVs,self.hdr.checksum) = struct.unpack('9I2H', sbuf)
						
					except:
						if self.dbg == True:
							print("(Header)Improper TLV structure found: ")
						return (False,v6,v7,v8)
					
					if disp == True:  
						self.headerShow()
					
					tlvCount = self.hdr.numTLVs
					if self.hdr.numTLVs == 0:
						return (True,v6,v7,v8)
						
					if self.sm == True:
						print("(Header)")
						
					sbuf = b""
					idx = 0
					lstate = 'TL'
					  
				elif idx > 44:
					idx = 0
					lstate = 'idle'
					return (False,v6,v7,v8)
					
			elif lstate == 'TL': #TLV Header type/length
				sbuf += ch
				idx += 1
				if idx == 8:
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						ttype,self.tlvLength = struct.unpack('2I', sbuf)
						#print("---{:d}: tlvCount({:d})-------ttype:tlvLength:{:d}:{:d}".format(self.hdr.numTLVs,tlvCount,ttype,self.tlvLength))
						if ttype not in typeList or self.tlvLength > 10000:
							if self.dbg == True:
								print("(TL)Improper TL Length(hex):(T){:d} (L){:x} numTLVs:{:d}".format(ttype,self.tlvLength,self.hdr.numTLVs))
							sbuf = b""
							idx = 0
							lstate = 'idle'
							self.port.flushInput()
							return (False,v6,v7,v8)
							
					except:
						if self.dbg == True:
							print("TL unpack Improper Data Found:")
						self.port.flushInput()
						return (False,v6,v7,v8)
					
					unitByteCount,lstate ,plen ,dataBytes,lenCount, numOfPoints = self.tlvTypeInfo(ttype,self.tlvLength,disp)
				
					if self.sm == True:
						print("(TL:{:d})=>({:})".format(tlvCount,lstate))
						
					tlvCount -= 1
					idx = 0  
					sbuf = b""
			
			elif lstate == 'V6-unit':
				sbuf += ch
				idx += 1
				if idx == unitByteCount :
					#print(":".join("{:02x}".format(c) for c in sbuf))
					#print("unitByte:{:d}".format(len(sbuf)))
					try:
						self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit = struct.unpack('5f', sbuf)
						#print("Unit  ==> elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit))
						sbuf = b""
						idx = 0
						lstate = 'V6'				
					except:
						if self.dbg == True:
							print("(6.0)Improper Type 6 unit Value structure found: ")
						return (False,v6,v7,v8)
					
					if self.sm == True:
						print("(V6-unit:{:d})=>({:})".format(tlvCount,lstate))
						#print("(V6-unit)=>({:})".format(lstate))
					
			elif lstate == 'V6': # count = Total Lentgh - 8
				sbuf += ch
				idx += 1
				if (idx%dataBytes == 0):
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						(e,a,d,r,s) = struct.unpack('2b3h', sbuf)
						elv = e * self.u.elevationUnit
						azi = a * self.u.azimuthUnit
						dop = d * self.u.dopplerUnit
						ran = r * self.u.rangeUnit
						snr = s * self.u.snrUnit
						#print("({:2d}:{:4d})(idx:({:4d}) elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(numOfPoints,lenCount,idx,elv,azi,dop,ran,snr))
						
						v6.append((elv,azi,dop,ran,snr))
						#print("point_cloud_2d.append:[{:d}]".format(len(point_cloud_2d)))
						sbuf = b""
					except:
						if self.dbg == True:
							print("(6.1)Improper Type 6 Value structure found: ")
						return (False,v6,v7,v8)
					
				if idx == lenCount:
					if disp == True:
						print("v6[{:d}]".format(len(v6)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V6:{:d})=>(idle) :true".format(tlvCount))
						return (True,v6,v7,v8)
						
					else: # Go to TL to get others type value
						lstate = 'TL' #'tlTL'
						if self.sm == True:
							print("(V6:{:d})=>(TL)".format(tlvCount))
					
				elif idx > lenCount:
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return (False,v6,v7,v8)
				
			elif lstate == 'V7':
				sbuf += ch
				idx += 1
				if (idx%dataBytes == 0):
					#print("V7:dataBytes({:d}) lenCount({:d}) index:{:d}".format(dataBytes,lenCount,idx))
					try:
						(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ) = struct.unpack('I9f', sbuf)
						v7.append((tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ))
						sbuf = b""
					except:
						if self.dbg == True:
							print("(7)Improper Type 7 Value structure found: ")
						return (False,v6,v7,v8)
						
				if idx >= lenCount:
					if disp == True:
						print("v7[{:d}]".format(len(v7)))
					 
					sbuf = b""
					if tlvCount == 0:
						lstate = 'idle'
						if self.sm == True:
							print("(V7)=>(idle) :true")		 
						return (True,v6,v7,v8)	
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						idx = 0
						if self.sm == True:
							print("(V7)=>(TL)")

				if idx > lenCount:
					idx = 0 
					lstate = 'idle'
					sbuf = b""
					return (False,v6,v7,v8)
				
			elif lstate == 'V8':
				idx += 1
				v8.append(ord(ch))
				if idx == lenCount:
					if disp == True:
						print("=====V8 End====")						
					sbuf = b""
					idx = 0
					lstate = 'idle'
					if self.sm == True:
						print("(V8:{:d})=>(idle)".format(tlvCount))
					return (True,v6,v7,v8)
				
				if idx > lenCount:
					sbuf = b""
					idx = 0
					lstate = 'idle'
					return (False,v6,v7,v8)







# Long range People Detect Raw Data (LPD)
# ver:0.0.1
# 2019/12/02
# parsing Long range People Detect 
# hardware:(Batman-201): ISK IWR6843 ES2.0
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: objPoint: [(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,state),....]
# v0.0.1 : 2019/12/06 release

import serial
import time
import struct
from dataclasses import dataclass

#import numpy as np

@dataclass
class objPoint:
	tid: int = 0
	x: float = 0.0
	y: float = 0.0
	z: float = 0.0
	vx: float = 0.0
	vy: float = 0.0
	vz : float = 0.0
	accX : float = 0.0
	accY : float = 0.0
	accZ : float = 0.0
	state : int = 0

@dataclass
class objSets:
	frameNum: int
	numObjs: int
	op: [objPoint]

#
# input data: { 32 bytes },{ 32 bytes }...{ 32 bytes };
# ; is end of data
# , is continue data
#

class LpdISK_kv:
	#				{        }        ,      ;
	magicWord =  [b'\x7B',b'\x7D',b'\x2C',b'\x3B']
	port = ""
	clen = 47 #exclude: delimeter
	
	def __init__(self,port):
		self.port = port
		print("(jb)Long range People Detect(LPD) lib initial")
		print("(jb)For Hardware:Batman-201(ISK)")
		print("(jb)Hardware: IWR-6843 ES2.0")
		print("(jb)Firmware: LPD")
		print("(jb)UART Baud Rate:921600")
		print("(jb)Data type: kv (key/Value)")
		print("==============Info=================")
		print("Output: [(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,state),....]")
		print("state = 0x00, N/A")
		print("===================================")
	
	def lpdRead(self,disp):
		ops = []
		frameNum = 0
		objMax = 0
		idx = 0
		lstate = 'idle'
		sbuf = b""
		while True:
			try:
				ch = self.port.read()
			except:
				print("(PC3D)---port.read() Exception---")
				return (False, [])
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[0]:
					#if disp:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					idx = 0
					lstate = 'iData'
					ops = [] #Object Point Set
					sbuf = ch
					if disp:
						print("-----------------------")
						print("(jb) idle-> idata")
				else:
					idx = 0
					sbuf = b""
					ops = [] 
					return (False,[])
					
			elif lstate == 'dataConti':
				if ch == self.magicWord[0]:
					sbuf = b""
					sbuf += ch
					idx = 0
					lstate = 'iData'
					if disp:
						print("(jb) dataConti-> iData: {}".format(sbuf))
				else:
					lstate = 'idle'
					if disp:
						print("(jb) dataConti-> idle")
					return (False,[])
		
			elif lstate == 'iData':
				sbuf += ch
				#print(":".join("{:02x}".format(c) for c in sbuf))  
				idx += 1
				if idx == self.clen and self.magicWord[1] == ch: #idxx = 31 + 12(ax,ay,az) + 3(fr) + 1 check=0x7d
					if disp:
						print("------rx data-----")
						print(":".join("{:02x}".format(c) for c in sbuf))  
					
					try: 
						#(h,f,fr,om,oi,tid,x,y,z,vx,vy,vz,ax,ay,az,state,c,t) = struct.unpack('<6c9f3c', sbuf)
						(h,f,fr,om,oi,tid,x,y,z,vx,vy,vz,ax,ay,az,state,c,t) = struct.unpack('<2cL3c9f3c', sbuf)
						op = objPoint(ord(tid),x,y,z,vx,vy,vz,ax,ay,az,ord(state))
						objMax = ord(om)
						frameNum = fr
						if disp:
							print("flowNum:{:d} frameNum:{:d} objMax:{:d} objIdx:{:d} tid:{:d}".format(ord(f)-ord('0'),frameNum,objMax,ord(oi),ord(tid)))
						lstate = 'eodCheck'
						ops.append(op)
						
						if objMax == 0:
							idx = 0
							lstate = 'idle'
							if disp:
								print("(jb)iData(0) -> idle state")
							return (True, objSets(frameNum,objMax,ops))
							
						sbuf = b""
						idx = 0
						if disp:
							print("(jb)iData(1) -> eodCheck state")
					except:
						lstate = 'idle'
						self.ops = []
						if disp:
							print("(PC3D)---iData Exception---")
					
				elif idx > self.clen + 3: #34 - 31= 3 -> 31+13+3 + [3] = 50
					lstate = 'idle'
					sbuf = b""
					print("*********data over*********")
					if disp:
						print("data over {:d} back to idle state".format(idx))
					frameNum = 0
					objMax = 0
					idx = 0
					ops = []
					return (False,[]) 
					
			elif lstate == 'eodCheck':
					if self.magicWord[2] == ch: # check ,
						lstate = 'dataConti'
						if disp:
							print("(jb(,))eodCheck-> dataCont (contiune read): length op count:{:d}".format(len(ops)))
					
					if self.magicWord[3] == ch: # check ;
						lstate = 'idle'
						idx = 0
						if disp:
							print("(jb(;))eodCheck->idle: length op count:{:d}".format(len(ops)))
						return (True, objSets(frameNum,objMax,ops))




# People Counting 3D OOB
# 2022/03/01: alpha
# 
# parsing People Counting 3D + OOB  
# hardware: BM-601 
#     
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
# v0.0.3
#===========================================
# output: V1,V2,V3,V4,V5,V6,V7 and V9 
# 

import serial
import time
import struct
import pandas as pd
import numpy as np

class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	timeInCPUcycle = 0
	numDetectedObj = 0
	numTLVs = 0
	subframeNumber = 0
	
class statistics:
	interFrameProcessTime = 0.0
	transmitOutTime = 0.0
	interFrameProcessMargin = 0.0
	interChirpProcessMargin = 0.0
	activeFrameCPULoad = 0.0
	interFrameCPULoad = 0.0
	
class temperatureStatistics:
	tempReportValid  = 0
	timeRadarSS  	 = 0
	tmpRx0Sens 		 = 0
	tmpRx1Sens		 = 0
	tmpRx2Sens		 = 0
	tmpRx3Sens		 = 0
	tmpTx0Sens		 = 0
	tmpTx1Sens		 = 0
	tmpTx2Sens		 = 0
	tmpPmSens 		 = 0
	tmpDig0Sens 	 = 0
	tmpDig1Sens 	 = 0
	 
	 
class unitS:
	elevationUnit:float = 0.0
	azimuthUnit : float = 0.0
	dopplerUnit :float = 0.0
	rangeUnit :float = 0.0
	snrUnit :float = 0.0

class Pc3_OOB:
	
	Q9 = 1.0 / 2**9 
	#magicWord =  [b'\x01',b'\x02',b'\x03',b'\x04',b'\x05',b'\x06',b'\x07',b'\x08',b'\0x99']
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	typeList = [1,2,3,4,5,6,7,9] 
	
	v2 = []
	v2df = []
	
	port = ""
	hdr = header
	u = unitS
	ss = statistics
	tss = temperatureStatistics
	
	# provide csv file dataframe
	# real-time 
	v1_col_names_rt = ['fN','type','X','Y','Z','doppler']
	v6_col_names_rt = ['fN','type','iFPT','xOutputT','iFPM','iCPM','aFCpu','iFCpu']
	v9_col_names_rt  = ['fN','type','tRValid','time','tmpRX0','tmpRX1','tmpRX2','tmpRX3','tmpTX0','tmpTX1','tmpTX2','tmpPm','tmpDig0','tmpDig1']
	
	 
	# read from file for trace point clouds
	fileName = ''
	
	sim_startFN = 0
	sim_stopFN  = 0 
	
	# add for interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm  = False #Observed StateMachine: True Show message
	plen = 16 
	
	def __init__(self,port):
		self.port = port
		print("\n\n\n(jb)mmWave OOB initial")
		print("(jb)version:v0.1.0")
		print("(jb)For Hardware:BM-601/BM-501")
		print("(jb)Hardware: BM601/BM501")
		print("(jb)Firmware:")
		print("(jb)UART Baud Rate:921600")
		print("(jb)Output: V1,V2,V3,V4,V5,V6,V7 & V9 data:(RAW)\n")
		print('v1: Detected Points\t\nv2: Range Profile\t\nv3: Noise Floor Profile\t\nv4: Azimuth Static Heatmap\t\nv5: Range-Doppler Heatmap\t\nv6: Statistics\t\nv7: Side Info for Detected Points\t\nv9: Temperature Statistic')
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	
	def headerShow(self):
		print("******* Header ********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("time in CPU cycles:\t{:d} us".format(self.hdr.timeInCPUcycle))
		print("number of Detected object:\t{:d}".format(self.hdr.numDetectedObj))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("***End Of Header***") 
			
			
			
	#for class internal use
	def tlvTypeInfo(self,dtype,count,dShow):
		
		sbyte = 8  #tlvHeader Struct = 8 bytes
		unitByte = 20 
		dataByte = 2
		pString = ""
		nString = "numOfPoints :"
		stateString = "V6-unit"
		retCnt = count
		nPoint = 0
		
		if dtype == 1: 
			unitByte = 0  
			dataByte= 16 #point bytes= 16bytes (X(4),Y(4),Z(4),doppler(4))
			pString = "Detected Points"
			stateString = "V1"
			nPoint = count/dataByte
			retCnt = count
			 
		elif dtype == 2: 
			unitByte = 0  
			dataByte= 2 #point bytes= 2bytes 
			pString = "Range Profile"
			stateString = "V2"
			nPoint = count/dataByte
			retCnt = count
			
		elif dtype == 3: 
			unitByte = 0  
			dataByte= 2 #point bytes= 2bytes
			pString = "Noise Floor Profile"
			nString = "1 set data"
			stateString = "V3"
			nPoint = count/dataByte
			retCnt = count
			
		elif dtype == 4: 
			unitByte = 0  
			dataByte= 2 #point bytes= 4bytes
			pString = "Azimuth Static Heatmap"
			nString = "1 set data"
			stateString = "V4"
			nPoint = count/dataByte
			retCnt = count
		
		elif dtype == 5: 
			unitByte = 0  
			dataByte= 2 #point bytes= 2bytes  (Range FFT size) * (Doppler FFT size) * 2 bytes (size of uint16_t)
			pString = "Range-Doppler Heatmap"
			nString = "1 set data"
			stateString = "V5"
			nPoint = count/dataByte
			retCnt = count
			
		elif dtype == 6:
			unitByte = 0   #pointUnit= 0bytes 
			sbyte = 8 	   #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte = 24  # 6 * 4 bytes
			pString = "Statistics"
			nString = "1 set data"
			nPoint = count/dataByte
			retCnt = count
			stateString = "V6"
		
		elif dtype == 7:  
			unitByte = 0  #pointUnit= 0
			sbyte = 8      #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 4    #pointStruct 8bytes:(snr(2),noise(2))
			pString = "Side Info for Detected Points"
			nPoint = count/dataByte
			retCnt = count
			stateString = "V7"
			
		elif dtype == 9: 
			unitByte = 0  
			sbyte = 8 #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 28 # 28 Bytes
			pString = "Temperature Statistics"
			stateString = "V9"
			nPoint = count/dataByte
			retCnt = count

		else:
			unitByte = 0
			sbyte = 1
			pString = "*** Type Error ***"
			stateString = 'idle'
		 

		#dShow = True
		if dShow == True:
			print("-----[{:}] ----".format(pString))
			print("tlv Type({:2d}Bytes):  \tv{:d}".format(sbyte,dtype))
			print("tlv length:      \t{:d}".format(count)) 
			print("{:}      \t{:d}".format(nString,int(nPoint)))
			print("value length:    \t{:d}".format(retCnt))
			print("StateString= {}".format(stateString))  
		
		return unitByte,stateString, sbyte, dataByte,retCnt, int(nPoint)
		
	def list2df(self,dck,l1,l6,l9):
		ll1 = pd.DataFrame(l1,columns=self.v1_col_names_rt)
		ll6 = pd.DataFrame(l6,columns=self.v6_col_names_rt) 
		ll9 = pd.DataFrame(l9,columns=self.v9_col_names_rt) 
		 
		return (dck,ll1,ll6,ll9)

	
#
# TLV: Type-Length-Value
# read TLV data
# Usuage:
# (dck,v1,v6,v9)  = radar.tlvRead(True,df = 'DataFrame')
# input:
#     disp: True:print message
#	  False: hide printing message
#  
# output:(return parameter)
# (pass_fail, v1, v6 & V9)
#  pass_fail: True: Data available    False: Data not available
#  v1: Detected Points
#  v2: Range Profile
#  v3: Noise Floor Profile
#  v4: Azimuth Static Heatmap
#  v5: Range-Doppler Heatmap
#  v6: Statistics
#  v7: Side Info for Detected Points
#  v8: Azimuth/Elevation Static Heatmap
#  v9: Temperature Statistic
#   
	def tlvRead(self,disp,df = None):
		
		
		idx = 0
		lstate = 'idle'
		sbuf = b""
		lenCount = 0
		unitByteCount = 0
		escapeCount = 0
		dataBytes = 0
		numOfPoints = 0
		tlvCount = 0
		pbyte = 16
		v1 = ([])
		self.v2 = ([])
		v3 = ([])
		v4 = ([])
		v5 = ([])
		v6 = ([])
		v7 = ([])
		v9 = ([])
		
		v1df = ([])
		self.v2df = ([])
		v3df = ([])
		v4df = ([])
		v5df = ([])
		v6df = ([])
		v7df = ([])
		v9df = ([])
		 
		
		while True:
			try:
				ch = self.port.read()
			except:
				return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
				
			#print(str(ch))
			if lstate == 'idle':
				#print(self.magicWord)
				if ch == self.magicWord[idx]:
					#print("*** magicWord:"+ "{:02x}".format(ord(ch)) + ":" + str(idx))
					idx += 1
					if idx == 8:
						idx = 0
						lstate = 'header'
						rangeProfile = b""
						sbuf = b""
						self.escapeCount = 0
				else:
					#print("not: magicWord state:")
					idx = 0
					rangeProfile = b""
					return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
		
			elif lstate == 'header':
				sbuf += ch
				idx += 1
				escapeCount += 1
				if idx == 32: 
					#print("------header-----")
					#print(":".join("{:02x}".format(c) for c in sbuf)) 	 
					#print("len:{:d}".format(len(sbuf))) 
					# [header - Magicword]
					
					try: 
						(self.hdr.version,self.hdr.totalPackLen,
						self.hdr.platform,self.hdr.frameNumber,
						self.hdr.timeInCPUcycle,self.hdr.numDetectedObj,
						self.hdr.numTLVs,self.hdr.subframeNumber) = struct.unpack('8I', sbuf)
						self.frameNumber = self.hdr.frameNumber
					except:
						if self.dbg == True:
							print("(Header)Improper TLV structure found: ")
						return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
					
					if disp == True:  
						self.headerShow()
					
					tlvCount = self.hdr.numTLVs
					#print("tlvCount:{:}".format(tlvCount))
					if self.hdr.numTLVs == 0:
						return self.list2df(True,v1df,v6df,v9df) if (df == 'DataFrame') else (True,v1,v6,v9)
						
					if self.sm == True:
						print("(Header)")
						
					sbuf = b""
					idx = 0
					lstate = 'TL'
					  
				elif idx > 64: #44
					idx = 0
					sbuf = b''
					lstate = 'idle'
					return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
					
			elif lstate == 'TL': #TLV Header type/length
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				
				if idx == 8:
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						ttype,self.tlvLength = struct.unpack('2I', sbuf)
						if self.dbg == True:
							print("(TL)--tlvNum:{:d}: tlvCount({:d})-------ttype:tlvLength:{:d}:{:d}".format(self.hdr.numTLVs,tlvCount,ttype,self.tlvLength))
						if ttype not in self.typeList or self.tlvLength > 20000:
							if self.dbg == True:
								print("(TL)Improper TL Length(hex):(T){:d} (L){:x} numTLVs:{:d}".format(ttype,self.tlvLength,self.hdr.numTLVs))
							sbuf = b""
							idx = 0
							lstate = 'idle'
							self.port.flushInput()
							return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
							
					except:
						if self.dbg == True:
							print("TL unpack Improper Data Found:")
						self.port.flushInput()
						return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
						
					
						
					unitByteCount,lstate ,plen ,dataBytes,lenCount, numOfPoints = self.tlvTypeInfo(ttype,self.tlvLength,disp)
					if self.dbg == True:
						print("unitByteCount={}  lstate ={} , bytes(type+len)={} dataBytes= {} lenCount ={} numOfPoint={}".format(unitByteCount,lstate ,plen ,dataBytes,lenCount, numOfPoints))
					#if ttype == 10:
					#	print("--pointCloud:((tlvLength({:d})-pointUnit(20)-tlvStruct(8))/8={:d}".format(self.tlvLength,numOfPoints))
					if self.sm == True:
						print("(TL:tlvCount:{:d})=>({:})".format(tlvCount,lstate))
					
					tlvCount -= 1
					#print("(TL->type:{:})--tlvNum:{:d}: tlvCount({:d})".format(ttype,self.hdr.numTLVs,tlvCount))
					idx = 0  
					sbuf = b""
					
					if ttype == 1 and self.tlvLength == 0 and tlvCount != 0:
						if self.dbg == True:
							print("========type=1 lenth = 0 ===========================================")
						idx = 0
						lstate = 'idle'
						
					if self.escapeCount > self.hdr.totalPackLen:
						if self.dbg == True:
							print("(ESC)========type={:} ========self.escapeCount > self.hdr.totoalPackLen==============================".format(lstate))
						idx = 0
						sbuf = b''
						lstate = 'idle'
					
			elif lstate == 'V1':
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				if (idx%dataBytes == 0):
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						(x,y,z,d) = struct.unpack('4f', sbuf)
						if self.dbg == True:
							print("({:2d}:{:4d})(idx:({:4d}) x:{:.4f} y:{:.4f} z:{:.4f} d:{:.4f}".format(numOfPoints,lenCount,idx,x,y,z,d))
						
						if (df == 'DataFrame'):
							v1df.append((self.hdr.frameNumber,'v1',x,y,z,d)) 
						else:
							v1.append((x,y,z,d))
						
						#print("point_cloud_2d.append:[{:d}]".format(len(point_cloud_2d)))
						sbuf = b""
						
					except:
						if self.dbg == True:
							print("(6.1)Improper Type 1 Value structure found: ")
						
						return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
				
				if idx == lenCount:
					if disp == True:
						print("v1[{:d}]".format(len(v1)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V1:{:d})=>(idle) :true".format(tlvCount))
						
						return self.list2df(True,v1df,v6df,v9df) if (df == 'DataFrame') else (True,v1,v6,v9)
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						if self.sm == True:
							print("(V1:{:d})=>(TL)".format(tlvCount))
					
				elif idx > lenCount or self.escapeCount > self.hdr.totalPackLen:
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
				
				
			
			elif lstate == 'V2' or lstate == 'V3' or lstate == 'V4' or lstate == 'V5' or lstate == 'V6' or lstate == 'V9':
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				#if (idx%dataBytes == 0):
				
				if lenCount == 0:
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
				
				if (idx%lenCount == 0):
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						if lstate == 'V2':
							rp = struct.unpack('256H', sbuf)
							rpf = [float(i) * self.Q9 for i in rp] 
							#print("V2:(Range Profile) (points:{:2d}:lenCount:{:4d})(idx:({:4d}) len(rpf)= {:}".format(numOfPoints,lenCount,idx,len(rpf)))
							self.v2 = rpf
							'''
							self.v2.append(rpf)
							
							if (df == 'DataFrame'):
								self.v2df.append((self.hdr.frameNumber,'v2',rpf)) 
							else:
								self.v2 = rpf
							'''
						elif lstate == 'V3':
							nf = struct.unpack('256H', sbuf)
							nff = [float(i) * self.Q9 for i in nf] 
							#print("V3:(Noise Floor Profile ({:2d}:{:4d})(idx:({:4d}) len(nff)= {:}".format(numOfPoints,lenCount,idx,len(nff)))
							 
							if (df == 'DataFrame'):
								v3df.append((self.hdr.frameNumber,'v3',nff)) 
							else:
								v3.append(nff)
						elif lstate == 'V4':
							ash = struct.unpack('4096h', sbuf)
							if self.dbg == True:
								print("V4:(Azimuth Static HeatMap ({:2d}:{:4d})(idx:({:4d}) len(azimuth-static ash)= {:}".format(numOfPoints,lenCount,idx,len(ash)))
							if (df == 'DataFrame'):
								v4df.append((self.hdr.frameNumber,'v4',ash)) 
							else:
								v4.append(ash)
								
						elif lstate == 'V5':
							rdh = struct.unpack('4096h', sbuf)
							rdhf = [float(i) * self.Q9 for i in rdh]
							if self.dbg == True: 
								print("V5:(Range-Doppler HeatMap ({:2d}:{:4d})(idx:({:4d}) len(range-doppler rd)= {:}".format(numOfPoints,lenCount,idx,len(rdhf)))
							#print(rdhf)
							if (df == 'DataFrame'):
								v5df.append((self.hdr.frameNumber,'v5',rdhf)) 
							else:
								v5.append(rdh)
								
						elif lstate == 'V6':
							#print("V6----lenCount:{:}".format(lenCount)) 
							
							(self.ss.interFrameProcessTime,  self.ss.transmitOutTime,
							self.ss.interFrameProcessMargin,self.ss.interChirpProcessMargin,
							self.ss.activeFrameCPULoad,      self.ss.interFrameProcessTime) = struct.unpack('6I', sbuf)
							#print("v6 usec:{:}".format(self.ss.interFrameProcessTime)) 
							
							if (df == 'DataFrame'):
								v6df.append((self.hdr.frameNumber,'v6',self.ss.interFrameProcessTime,  
								self.ss.transmitOutTime, self.ss.interFrameProcessMargin,
								self.ss.interChirpProcessMargin, self.ss.activeFrameCPULoad, self.ss.interFrameProcessTime) )
							else:
								v6.append((self.ss.interFrameProcessTime,  
								self.ss.transmitOutTime, self.ss.interFrameProcessMargin,
								self.ss.interChirpProcessMargin, self.ss.activeFrameCPULoad, self.ss.interFrameProcessTime))
						
						elif lstate == 'V9':
							#print("V9----lenCount:{:}".format(lenCount)) 
							
							(self.tss.tempReportValid,self.tss.timeRadarSS,
							self.tss.tmpRx0Sens,self.tss.tmpRx1Sens,self.tss.tmpRx2Sens,self.tss.tmpRx3Sens,
							self.tss.tmpTx0Sens,self.tss.tmpTx1Sens,self.tss.tmpTx2Sens,
							self.tss.tmpPmSens,
							self.tss.tmpDig0Sens,self.tss.tmpDig1Sens ) = struct.unpack('2I10H', sbuf)
							
							if (df == 'DataFrame'):
								v9df.append((self.hdr.frameNumber,'v9',self.tss.tempReportValid,self.tss.timeRadarSS,
								self.tss.tmpRx0Sens,self.tss.tmpRx1Sens,self.tss.tmpRx2Sens,self.tss.tmpRx3Sens,
								self.tss.tmpTx0Sens,self.tss.tmpTx1Sens,self.tss.tmpTx2Sens,
								self.tss.tmpPmSens,
								self.tss.tmpDig0Sens,self.tss.tmpDig1Sens )) 
							else:
								v9.append((self.tss.tempReportValid,self.tss.timeRadarSS,
								self.tss.tmpRx0Sens,self.tss.tmpRx1Sens,self.tss.tmpRx2Sens,self.tss.tmpRx3Sens,
								self.tss.tmpTx0Sens,self.tss.tmpTx1Sens,self.tss.tmpTx2Sens,
								self.tss.tmpPmSens,
								self.tss.tmpDig0Sens,self.tss.tmpDig1Sens))
								
							
							
						
						#print("point_cloud_2d.append:[{:d}]".format(len(point_cloud_2d)))
						sbuf = b""
						#lstate = 'TL'
						
					except:
						if self.dbg == True:
							print("Improper Type {} Value structure found: ".format(lstate))
						return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
				
				if idx == lenCount:
					'''
					if disp == True:
						if lstate == 'V2':
							print("v2[{:d}]".format(len(v2)))
						else:
							print("v3[{:d}]".format(len(v3)))
					'''
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						if self.sm == True:
							print("({}:tlvCount({:d})) lenCount:{}=>(idle)".format(lstate,tlvCount,lenCount))
						lstate = 'idle'
						
						return self.list2df(True,v1df,v6df,v9df) if (df == 'DataFrame') else (True,v1,v6,v9)
						
					else: # Go to TL to get others type value
						if self.sm == True:
							print("({}:tlvCount({:d})) lenCount:{}=>(TL)".format(lstate,tlvCount,lenCount))
						lstate = 'TL'
											
				elif idx > lenCount or self.escapeCount > self.hdr.totalPackLen:
					print("(ESC)========type={:} lenth = 0 =============self.escapeCount > self.hdr.totoalPackLen==============================".format(lstate))
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
				
					
			elif lstate == 'V7': 
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				if (idx%dataBytes == 0):
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						(snr,noise) = struct.unpack('2H', sbuf)
						v7.append((snr,noise))
						sbuf = b""
						
					except:
						if self.dbg == True:
							print("(6.1)Improper Type 6 Value structure found: ")
						
						return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
					
				if idx == lenCount:
					if disp == True:
						print("v7[len:{:d}]".format(len(v7)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V7:{:d})=>(idle) :true".format(tlvCount))
						
						return self.list2df(True,v1df,v6df,v9df) if (df == 'DataFrame') else (True,v1,v6,v9)
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						if self.sm == True:
							print("(V7:{:d})=>(TL)".format(tlvCount))
							
					
				elif idx > lenCount or self.escapeCount > self.hdr.totalPackLen:
					print("(ESC)========type={:} lenth = 0 =============self.escapeCount > self.hdr.totoalPackLen==============================".format(lstate))
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v6df,v9df) if (df == 'DataFrame') else (False,v1,v6,v9)
					
		
				
	def v67Simlog(frameNum):
		global sim_startFN,sim_stopFN
		s_fn = frameNum + sim_startFN
		#print("frame number:{:}".format(s_fn))
		v6d = v6sim[v6sim['fN'] == s_fn]
		#v6d =  v6dd[v6dd['doppler'] < 0.0]
		#print(v6d)
		v7d = v7sim[v7sim['fN'] == s_fn]
		chk = 0
		if v6d.count != 0:
			chk = 1
		return (chk,v6d,v7d)
		
	def getRecordData(self,frameNum):
		s_fn = frameNum + self.sim_startFN
		#print("frame number:{:}".format(s_fn))
		v6d = self.v6simo[self.v6simo['fN'] == s_fn]
		#v6d =  v6dd[v6dd['doppler'] < 0.0]
		#print(v6d)
		v7d = self.v7simo[self.v7simo['fN'] == s_fn]
		v8d = self.v8simo[self.v8simo['fN'] == s_fn]
		chk = 0
		if v6d.count != 0:
			chk = 1
		return (chk,v6d,v7d,v8d)
		
	def readFile(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		#          ['time','fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz']
		df = pd.read_csv(self.fileName, names = self.v6_col_names, skiprows = [0,11,12]) 
		df.dropna()
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		print(df.info())
		print(df.info(memory_usage="deep")) 
		
		v6simOri = df[(df.type == 'v6')]
		#print("-------------------v6sim------------:{:}".format(v6simOri.shape))
		#self.v6simo = v6simOri.loc[:,['fN','elv','azimuth','range','doppler','snr']]
		self.v6simo = v6simOri.loc[:,['fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz']]
		self.v6simo['elv'] = self.v6simo['elv'].astype(float, errors = 'raise') 
		
		df7 = pd.read_csv(self.fileName, names = self.v7_col_names, skiprows = [0])  
		
		v7simc = df7[df7['type'] == 'v7']
		self.v7simo  = v7simc.loc[:,['fN','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi','tid']]
		self.sim_startFN = df['fN'].values[0]
		self.sim_stopFN  = df['fN'].values[-1]
		'''
		print("---------start frame number---------:{:}".format(self.sim_startFN))
		print("---------stop  frame number---------:{:}".format(self.sim_stopFN))
		print(self.v7simo)
		print("-----v8sim data lib-----")
		'''
		v8simc = df[df['type'] == 'v8']
		self.v8simo  = v8simc.loc[:,['fN','type','elv']]
		self.v8simo.columns = ['fN','type','targetID']
		#print(self.v8simo)
		return (self.v6simo,self.v7simo,self.v8simo)


